(* a language for defining CUDA kernel *)

open Config

exception Build_error of string

type kind =
  (* a set of operation types over parallel arrays *)
  | Initialize
  | Map
  | Reduce
  | UnitlessReduce                      (* reduce without a zero value *)

let cxx_suffix : string = "cu"
let backend_name : string = "cuda"

let string_of_kind : kind -> string =
  function
  | Initialize     -> "initialize"
  | Map            -> "map"
  | Reduce         -> "reduce"
  | UnitlessReduce -> "unitless_reduce"

let element_type_of : Ikra_type.t -> Ikra_type.t =
  let open Ikra_type in
      function
        | Tarray(_, typ) -> typ
        | t -> invalid_arg ("expected an array type, but "
                            ^ (Ikra_type.string_of t))

type kernel = 
    (* a set of exceprted information from the source code that is
       needed for generating kernel functions and their wrappers. *)
    {
      kind : kind;
      name : string;
      free_vars : Ast.tast list;
      params : Ast.parameter_pattern;
      type_decls : Ast.type_decl list;
      body : Ast.tast list;
      body_type : Ikra_type.t;
      receiver : Ast.tast;
      top_decls : Ast.tast;
      return_type : Ikra_type.t;
    }

let top_decls : kernel list -> Ast.tast list =
  let open Ast in
  let rec loop : Ast.tast -> Ast.tast list = fun (e,t) -> match e with
    | Method_def _ | Constant_assign _
      -> [e,t]
    | Block(exps) | Type_decl(_,exps)
      -> List.concat (List.map loop exps)

    | Proc _
    | Call _ | Literal _ | Constant _ | Local_assign _ | Local_var _
    | Yield _ | Array _ | While _ | If _ | For _ | Return _
      -> [] in
  fun kernels -> 
    List.concat(List.map (fun {top_decls=t} -> loop t) kernels)


let fv_name : Ast.tast -> string =
(* fv_name fv extracts the variable name of a free variable *)

  function
    | Ast.Local_var v, _ -> v
    | _ -> invalid_arg "unexpected type of node"


type var_iname_type_cname = Ast.tast * string

type env = {
  wrapper_name : string;
  kernel_fun_name : string;
  body_fun_name : string;

  argspecs : var_iname_type_cname list;

  body_ctype : Cuda.cuda_type;
  fvspecs : (Ast.tast * string) list;

  body_args : (Cuda.cuda_type * Cuda.variable) list;
  element_statements : Cuda.statement list;

  (* input_array_itype : Ikra_type.t; *)
  (* output_array_itype : Ikra_type.t; *)
  (* return_itype : Ikra_type.t; *)
  (* input_element_type : Cuda.cuda_type; *)
}

let unwrap : var_iname_type_cname -> Cuda.exp -> Cuda.exp =
  (* (unwrap (ikra_var_exp,C_var_name) C_var_exp) generates a C
     expression that unwraps the given variable of a Ruby value into
     a C value. *)
  fun ((_,typ),name) -> (match typ with
    | Ikra_type.Tint -> Cuda_constructors._NUM2INT
    | Ikra_type.Tfloat -> Cuda_constructors._IKRA_NUM2FLOAT
    | Ikra_type.Tarray(_,element_type) ->
      fun exp -> Cuda_constructors.get_parray
	(exp, Cuda.cuda_type_of_ikra_type element_type)
    (* TODO: support tuples, or do something not directly unwrap
       tuples here. *)
    | _ -> invalid_arg ("cannot unwrap "
                        ^ (Ikra_type.string_of typ) ^ " for " ^ name))
    
let make_var : (Ikra_type.t * string) -> var_iname_type_cname =
  fun (typ,name) -> ((Ast.Local_var name, typ), name)

let wrapper_arg : var_iname_type_cname -> (Cuda.cuda_type * string) =
  fun ((ivar,typ),cname) -> (Cuda_constructors._VALUE,cname)

let loop_arg : var_iname_type_cname -> (Cuda.cuda_type * string) =
  fun ((ivar,typ),cname) ->
    try (Cuda.cuda_type_of_ikra_type typ, cname)
    with Invalid_argument(msg) -> invalid_arg(msg ^ " for " ^ cname)

let wrap_c_value : Cuda.cuda_type -> Cuda.exp -> Cuda.exp =
  let open Ikra_type in
      let open Cuda_constructors in
          function
            | Cuda.CUDAint -> rb_int_new
            | Cuda.CUDAfloat -> rb_float_new
            | _ -> (fun _ -> invalid_arg "UNSUPPORTED_UNWRAPPER") 

let rec zero_to = function
  | 0 -> []
  | n when n < 0 -> invalid_arg "expects a non-negative number"
  | n -> (zero_to (n-1)) @ [n-1]

let rec flatten_tuples
    : var_iname_type_cname list -> var_iname_type_cname list =

  (* to construct an identical list of the given parameter list
     for kernel except that each array of tuples is replaced with
     arrays of tuple elements, and that each array of neighbors
     is replaced with an array with a padding value.  *)

  let open Ikra_type in
  let open Ast in 
  fun argspecs ->
    List.flatten
      (List.map (function
        | (Local_var iname,Tarray(n,Ttuple typs)),cname ->
          flatten_tuples
            (List.map2 (fun i typ ->
              let is = "_t"^(string_of_int i) in
              ((Local_var (iname^is), Tarray(n,typ)),(cname^is))) 
               (zero_to (List.length typs))
               typs)
        | (Local_var iname, Tarray(n,Tneighbor(m,n',t))),cname
          when n=n' ->
          [(Local_var (iname ^ "_array"), Tarray(n,t)), cname ^ "_array";
           (Local_var (iname ^ "_pad"  ), t),           cname ^ "_pad";]
        | ((e, Tarray(_,Ttuple _)) as te),cname 

        | ((e, Tarray(_,Tneighbor(_,_,_))) as te),cname -> (* never happens *)
          invalid_arg ("no local var type for "^ (string_of te))

        | other -> [other])
         argspecs)

let wrapper_template : env -> kind -> Cuda.fundef list = fun e  -> 
  let open Cuda_constructors in function
    | Initialize ->
      let argspecs = e.argspecs @ e.fvspecs in
      [defun _VALUE e.wrapper_name
	((_VALUE,"self")::(List.map wrapper_arg argspecs))
        (function self::args ->
          [
            _exp(e.kernel_fun_name **@
                   (List.map2 unwrap argspecs args));
            return _Qnil;
          ] | _ -> [])]
    | Reduce  | UnitlessReduce ->
      let argspecs = e.argspecs @ e.fvspecs in
      [defun _VALUE e.wrapper_name
	((_VALUE,"self")::(List.map wrapper_arg argspecs))
        (function self::args ->
          [
            return (wrap_c_value e.body_ctype
                      (e.kernel_fun_name **@
			 (List.map2 unwrap  argspecs args)));
          ] | _ -> [])]
    | Map ->
      let argspecs = e.argspecs @ e.fvspecs in
      let argspecs = flatten_tuples argspecs in
      [defun _VALUE e.wrapper_name
	((_VALUE,"self")::(List.map wrapper_arg argspecs))
        (function self::args ->
          [
            _exp(e.kernel_fun_name **@
		   (List.map2 unwrap argspecs args));
            return _Qnil;
          ] | _ -> [])]

(* TODO: this is moved to util.ml, change the call site to there *)
let rec split_last = function
  | [last] -> ([],last)
  | hd::tl -> let butlast, last = split_last tl in
	      hd::butlast, last
  | _ -> invalid_arg "empty"

let rec map_range : int -> int -> (int -> 'a) -> 'a list =
  fun min max f ->
    if min = max then [f max]
    else (f min)::(map_range (min+1) max f) 

let product =
  let open Cuda_constructors in
  List.fold_left (fun x y -> x *@ y) (_num 1)

let loop_template : env -> kind -> Cuda.fundef list = fun e ->
  let split0 argspecs generator args_fvs = 
    let rec loop args = function
      | (0,fvs) -> generator (List.rev args, fvs)
      | (n,arg::args_fvs) -> loop (arg::args) (n-1,args_fvs) 
      | (n,[]) -> invalid_arg
        (Printf.sprintf "too few args: expected=%s actual=%s"
         (String.concat "," (List.map (fun (_,iname) -> iname) argspecs))
         (String.concat "," (List.map Cuda.string_of_exp args_fvs))) in
    loop [] (List.length argspecs, args_fvs)  in

  let split = split0 e.argspecs in
  (* let split generator args_fvs = *)
  (*   let rec loop args = function *)
  (*     | (0,fvs) -> generator (List.rev args, fvs) *)
  (*     | (n,arg::args_fvs) -> loop (arg::args) (n-1,args_fvs)  *)
  (*     | (n,[]) -> invalid_arg "too few args" in *)
  (*   loop [] (List.length e.argspecs, args_fvs) in *)

  let rec splits arg_nums generator args_fvs = 
    let rec loop = function
      | [],[]  -> []
      | n::arg_nums, args_fvs ->
        (Util.take n args_fvs)::(loop (arg_nums, Util.drop n args_fvs))
      | [], _::_ -> invalid_arg "numbers and args do not match" in
    generator (loop (arg_nums,args_fvs)) in
    

  let loopgen : Cuda.exp list -> (Cuda.exp -> (Cuda.exp list) -> Cuda.statement)
    -> Cuda.statement =
    (* loopgen [size1;...;sizeN] (fun array_index_epr index_vars -> BODY_EXP) =
       for (int _i1 = 1; _i1 < size1; _i1++)
       ...
       for (int _iN = 0; _i1 < size0; _iN++)
       BODY_EXP[index_epr = ((_i1*size2 +_i2)*size3 + ... _iN-1)*sizeN
       index_vars = [_i1;...;_iN] ]
    *)

    (* TODO: separate a fragment that generates loop indices as it is
       not always needed. *)

    fun size_vars body_gen ->
      let open Cuda_constructors in 
      let rec loop dim index_exp index_vars = function
        | [] -> body_gen index_exp (List.rev index_vars)
        | size_var::size_vars ->
          _for _int (Printf.sprintf "_i%d" dim) 
               (fun index_var -> (_num 0, index_var <@ size_var, !++ index_var,
                 loop (succ dim)
                   ((index_exp *@ size_var) +@ index_var)
                   (index_var::index_vars)
		   size_vars)) in
      loop 0 (_num 0) [] size_vars in
      
      let open Cuda_constructors in
      let gen_index size_vars ?(else_clause=(fun _ -> (_block []))) gen =
            _let _int "$offset" ((blockIdx_x *@ blockDim_x) +@ threadIdx_x) 
              (fun offset  ->
               [ _if (offset <@ (product size_vars))
                  (_block
                  (let rec loop = function
                    | _,[_],index_vars -> gen offset (List.rev index_vars)
                    | dim,size_hd::size_tl,idxs ->
                      _let _int ("_i" ^ (string_of_int dim))
                        (offset %@ (size_hd *@ (product size_tl))
                         /@ (product size_tl)) (fun i ->
                           loop (succ dim, size_tl, i::idxs))
                    | _ -> raise Not_found in
                   loop (0,size_vars @ [_num 1],[])))
                  (else_clause offset)]) in
  let open Cuda_constructors in function
    | Initialize ->
      let loop_args =  (List.map loop_arg
				       (e.argspecs @ e.fvspecs)) in
      (* TODO: support construction of zipped arrays *)
      [__global__ void (e.body_fun_name ^ "$wrap") loop_args
          (split (function ptr::size_vars, fvs -> 
            (* TODO instead generating a loop, first compute the offset exp from thread ids, and then  *)
            gen_index size_vars (fun offset_exp index_vars -> [
              _exp((_aref ptr offset_exp) =@ (e.body_fun_name **@
					       (index_vars @ fvs)));
            ]) | _ -> []));
       inline void e.kernel_fun_name loop_args 
        (split (function ptr::size_vars, fvs -> 
        (_let _int "arraySize" (product size_vars) (fun arraySize ->
        [_if (arraySize <=@ (_num 0))
        (_block
          [(return_)]
        )
        (_block       
          (_kernel_call (e.body_fun_name ^ "$wrap")
            (((arraySize) +@
               _NUM_THREADS -@ (_num 1))/@_NUM_THREADS, _NUM_THREADS)
              (ptr::size_vars@fvs))
        )]))
	   | _ -> []));
      ]
    | Reduce -> 
      let loop_args =  (List.map loop_arg
         (e.argspecs @ e.fvspecs)) in
			let freevars = (List.map loop_arg e.fvspecs) in
      (* TODO: support zipped arrays *)
(*			usage of defun
let defun : ?modifier:string -> cuda_type -> string -> (cuda_type * string) list
  -> (exp list -> statement list) -> fundef = *)
      [ __device__inline e.body_ctype (e.kernel_fun_name^"warpReduce") ((e.body_ctype,"val")::(_int,"reduceSize")::freevars) (splits [1;1;(List.length freevars)] (function [[value];[reduceSize];fvs] ->
	(_let _int "lane" (threadIdx_x %@ (_num 32)) (fun lane ->
  (_let _int "offset" ((reduceSize +@ (_num 1)) /@ (_num 2))  (fun offset ->
  (_let0 e.body_ctype "val2" (fun value2 ->
  [
    (_while (reduceSize >@ (_num 1))
    (_block
    [
      (_exp (value2 =@ ("__shfl_down" **@ [value;offset])));
(*
   All of the threads in a warp should participate in the execution of __shfl_down(), because the
   receiver threads can't receive the value of the server threads.
*)
      (_if (((lane +@ offset)) <@ reduceSize)
        (_block
		    [
          (_exp (value =@ (e.body_fun_name **@ (value::value2::fvs))))
        ]
        )
        (_block []));
			(_exp (reduceSize =@ ((reduceSize +@ (_num 1)) /@ (_num 2))));
      (_exp (offset =@ ((reduceSize +@ (_num 1)) /@ (_num 2))));
    ]));
    (return value)
  ]))))))
  | _ -> []));
   __device__inline e.body_ctype (e.kernel_fun_name^"blockReduce") ((e.body_ctype,"val")::(_int,"reduceSize")::freevars) (splits [1;1;(List.length freevars)] (function [[value];[reduceSize];fvs] ->
  (_let0 (Cuda.CUDAstatic (Cuda.CUDAshared (Cuda.CUDApointer (e.body_ctype,[32])))) "shared" (fun shared ->
  (_let _int "lane" (threadIdx_x %@ (_num 32)) (fun lane ->
  (_let _int "widInBlock" (threadIdx_x /@ (_num 32)) (fun widInBlock ->
  [
  (_exp (value =@ ((e.kernel_fun_name^"warpReduce") **@ (value::(_ifexp (blockDim_x ==@ reduceSize) (_num 32) (reduceSize %@ (_num 32)))::fvs))));
(*
  If the condition of the _ifexp is true, all of the threads in the warp participate in the reduction,
  so 32(warp size) is passed to the warpReduce function, if false, some of the threads in the warp do,
  so (arraySize % 32) is passed.
*)
  (_exp ("__syncthreads" **@ []));  
  (_if (lane ==@ (_num 0))
  (_block
  [
    (_exp ((_aref shared widInBlock) =@ value))
  ]
  )
  (_block []));
  (_if (widInBlock ==@ (_num 0))
  (_block
  [
    (_exp (value =@ (_aref shared threadIdx_x)));
    (_exp (value =@ ((e.kernel_fun_name^"warpReduce") **@ (value::(((reduceSize +@ (_num 32)) -@ (_num 1)) /@ (_num 32))::fvs))))
  ]
  )
  (_block []));
  (return value)
  ])))))) | _ -> []));
    __global__ void (e.body_fun_name^"deviceReduceKernel") ((?* (e.body_ctype), "in")::(?* (e.body_ctype), "out")::(_int, "arraySize")::(e.body_ctype, "unit")::freevars) (splits [1;1;1;1;(List.length freevars)] (function [[arrayIn];[arrayOut];[arraySize];[unit];fvs] ->
  (_let _int "id" ((blockIdx_x *@ blockDim_x) +@ threadIdx_x) (fun id ->
  (_let _int "threadsPerGrid" (gridDim_x *@ blockDim_x) (fun threadsPerGrid ->
	(_let e.body_ctype "val" unit (fun value ->																											 
  [
  (_if (id <@ arraySize)
    (_block
    [
      (_exp (value =@ (_aref arrayIn id)))
    ]
    )
    (_block [])
  );
  (_for _int "_i" (fun _i -> ((id +@ threadsPerGrid), (_i <@ arraySize), (_i =@ (_i +@ (threadsPerGrid))),
  (_block
  [
    (_exp (value =@ (e.body_fun_name **@ (value::(_aref arrayIn _i)::fvs))))
  ]
  ))
  ));
  (_exp ("__syncthreads" **@ []));
  (_exp (value =@ ((e.kernel_fun_name^"blockReduce") **@ (value::(_ifexp ((blockIdx_x *@ blockDim_x) <@ arraySize) (blockDim_x) (arraySize %@ blockDim_x))::fvs))));
(*
  If the condition of the _ifexp is true, all of the threads in the block participate in the reduction,
  so blockDim_x(the number of threads in a block) is passed to the blockReduce function, if false, some of the threads in the block do,
  so (arraySize % blockDim_x) is passed.
*)
  (_if (threadIdx_x ==@ (_num 0))
    (_block
    [
      (_exp ((_aref arrayOut blockIdx_x) =@ value))
    ]
    )
    (_block [])
  )
  ])))))) | _ -> []));
  inline void (e.kernel_fun_name^"deviceReduce") ((?* (e.body_ctype), "in")::(?* (e.body_ctype), "out")::(_int, "arraySize")::(_int, "threads")::(_int, "blocks")::(e.body_ctype, "unit")::freevars) (splits [1;1;1;1;1;1;(List.length freevars)] (function [[arrayIn];[arrayOut];[arraySize];[threads];[blocks];[unit];fvs] ->
  (
  (_kernel_call (e.body_fun_name^"deviceReduceKernel") (blocks,threads) (arrayIn::arrayOut::arraySize::unit::fvs))@
  (_kernel_call (e.body_fun_name^"deviceReduceKernel") ((_num 1), threads) (arrayOut::arrayOut::blocks::unit::fvs))																							  ) | _ -> []));
  inline e.body_ctype e.kernel_fun_name loop_args
    (split (function ptr::unit::size_vars, fvs ->
    (_let _int "arraySize" (product size_vars) (fun arraySize ->
    [
    (_if (arraySize <=@ (_num 0))
    (_block
      [(return unit)]
    )
		(_block
      (_let0 ?*(e.body_ctype) "$tmp_array" (fun tmp_array ->
      (_let e.body_ctype "$accum" (product size_vars) (fun accum ->
      (_let _int "threads" (_num 256) (fun threads -> (* threads should be a multiple of 32*)
      (_let _int "warpsPerBlock" (threads /@ (_num 32)) (fun warpsPerBlock ->
      (_let _int "blocks" ("min" **@ [(((arraySize +@ threads) -@ (_num 1)) /@ threads);(((_num 64) /@ warpsPerBlock) *@ (_num 14))]) (fun blocks ->
(* 
  64 is the maximum number of warps which can be dealt with by a SMX (Stream Multiprocessor eXtreme)  
  14 is the number of SMXs in a GPU (K20x) 
  TODO: these number should be changed into variables
 *)
      [
      (_exp (cudaMalloc (!& tmp_array) (sizeof(e.body_ctype) *@ blocks)));
      (_exp ((e.kernel_fun_name^"deviceReduce") **@ (ptr::tmp_array::arraySize::threads::blocks::unit::fvs)));
      (_exp (cudaMemcpy (!& accum) tmp_array (sizeof e.body_ctype) CudaMemcpyDeviceToHost));
      (_exp (cudaFree tmp_array));
      (* (return (e.body_fun_name **@ [accum;unit])) *) (* reduce に渡された unit を 最後にreduce *)
      (return accum)
      ]
			))))))))))
    ))]))
     | _ -> []));
  ]
    | UnitlessReduce -> 
      (* TODO: support zipped arrays *)
      [inline e.body_ctype e.kernel_fun_name (List.map loop_arg
				       (e.argspecs @ e.fvspecs))

	(* The parameters are ordered so that The vairable number of
	   parameters come last. *)
        (split (function ptr::size_vars, fvs ->
          (* TODO: it is OK to use 0 as a dummy initial value with int
             and float accum, but not safe when we support other element
             values like tuples. *)
          _let e.body_ctype "accum" (_num 0) (fun accum -> [
	    loopgen size_vars (fun offset_exp index_vars ->
              _block
                (_let Cuda.CUDAint "_idx" offset_exp (fun idx -> [
                  _if(idx ==@ (_num 0))
                    (_exp(accum =@ _aref ptr idx))
                    (_exp(accum =@
                        (e.body_fun_name **@
                          (accum::_aref ptr idx::fvs))))
              ])));
            return accum;
          ]) | _ -> []))]
    | Map -> 
      (* since we support array of tuples, the arguments are
         treated in this way.  First, e.argspecs may contain an
         argument of type Array<n,Tuple<t1..tm>>.  This will be
         expanded to Array<n,t1>,...,Array<n,tm>.  We assume that the
         returned array do not contain tuples, and hence no expansion.

         Now e.argspecs == [input array, output array, size1,..,sizeN]
         and e.fvspecs == [freevars...].

         After expansion, argspecs == [input array1, ... input arrayM,
         output array, size1,..,sizeN, freevars...]

      *)
      match e.argspecs with
        | ia_spec::oa_spec::sizes ->
          let ia_specs = flatten_tuples [ia_spec] in
          let argspecs = ia_specs @ (oa_spec::(sizes @ e.fvspecs)) in
          let generate_aref ptrs indices sizes offset_exp =
            let open Ikra_type in
            match ia_spec, ptrs with
            (* TODO generalize to m-point neighbors *)
            | ((iname,Tarray (1,Tneighbor(3, (1 as rank), typ))),cname),
            [ptr;pad] ->
               map_range (-1) 1 (fun i ->
                 ("safe_aref" ^
                          (string_of_int rank)) **@ 
                   (List.concat [[ptr];
                                 List.map2 (+@) indices [_num i];
                                 sizes;[pad]]))
            | ((iname,Tarray (2,Tneighbor(9, (2 as rank), typ))),cname),
              [ptr;pad] ->
              List.flatten
                (map_range (-1) 1 (fun i ->
                  (map_range (-1) 1 (fun j ->
                    ("safe_aref" ^
                             (string_of_int rank)) **@ 
                      (List.concat [[ptr];
                                 List.map2 (+@) indices [_num i; _num j];
                                 sizes;[pad]])))))

            | ((iname,Tarray (3,Tneighbor(27, (3 as rank), typ))),cname),
              [ptr;pad] ->
              (* TODO:
                 
                 Currently, it generates the following driver code

                 output[i*size1*size2 + j*size2 + k] =
                 body(safe_aref(input,i-1,j-1,k-1,size0,size1,size2,pad),
                      safe_aref(input,i-1,j-1,k+0,size0,size1,size2,pad),
                      safe_aref(input,i-1,j-1,k+1,size0,size1,size2,pad),
                      safe_aref(input,i-1,j+0,k-1,size0,size1,size2,pad),
                      ...);

                 which is not efficint because safe_aref checks all
                 the indices.  Faster code is

                 int idx = i*size1*size2 + j*size2 + k;
                 c = input[idx];
                 output[i*size1*size2 + j*size2 + k] =
                 body(i==0||j==0||k==0 ? c : input[idx-size1*size2-size2-1],
                      i==0||j==0       ? c : input[idx-size1*size2-size2],
                      i==0||j==0||k=size2-1?c: input[idx-size1*size2-size2+1],
                      i==0||    ||k==0?c:      input[idx-size1*size2      -1],
                      i==0||          ?c:      input[idx-size1*size2],
                      i==0||    ||k=size2-1?c: input[idx-size1*size2      +1],
                      ...);
                 
                 This needs to change at the generation of loops as
                 well as the code here that constructs array accesses.

              *)
              List.flatten
                (map_range (-1) 1 (fun i ->
                  List.flatten
                    (map_range (-1) 1 (fun j ->
                      (map_range (-1) 1 (fun k ->
                    ("safe_aref" ^
                             (string_of_int rank)) **@ 
                      (List.concat [[ptr];
                                 List.map2 (+@) indices [_num i; _num j; _num k];
                                 sizes;[pad]])))))))

            | ((iname,Tarray _),cname),ptrs  ->
              (* TODO: rather than receiving offset_exp from the caller,
              just build offset_exp here. *)
              List.map (fun ptr -> _aref ptr offset_exp) ptrs
            | (e,cname), ptrs ->
              invalid_arg (Printf.sprintf "unexpected (%s,%s) [ptrs]"
                             (Ast.string_of ~show_type:true e) cname) in
            let loop_args = (List.map loop_arg argspecs) in
          [__global__ void (e.body_fun_name ^ "$wrap")
              loop_args
            (splits [List.length ia_specs; 1; List.length sizes;
                     List.length e.fvspecs]
               (function [ptrs;[ptr2];size_vars;fvs] ->
	         
                   gen_index size_vars (fun offset_exp index_vars -> [
	             _exp((_aref ptr2 offset_exp) =@
                         (e.body_fun_name **@
                           (  (* TODO support array of neighbors *)
                              (generate_aref ptrs index_vars size_vars
                              offset_exp) @
                              (* (List.map (fun ptr ->_aref ptr offset_exp) ptrs) *)
                               fvs)))]);
	          | _ -> []));
           inline void e.kernel_fun_name loop_args
             (splits [List.length ia_specs; 1; List.length sizes;
                      List.length e.fvspecs]
                (function ([ptrs;[ptr2];size_vars;fvs] as params) ->
                (_let _int "arraySize" (product size_vars) (fun arraySize ->
                [_if (arraySize <=@ (_num 0))
                (_block
                  [(return_)]
                )
                (_block       
                  (_kernel_call (e.body_fun_name ^ "$wrap")
                    (((arraySize) +@
                       _NUM_THREADS -@ (_num 1))/@_NUM_THREADS, _NUM_THREADS)
                      (List.flatten params))
                )]))
                  (* (List.append (ptr::size_vars) fvs) *)
                  | _ -> []));
          ]
        | _ -> invalid_arg "argspecs have unexpected number of arguments"

let ident_of_int : int -> string = function
  | n when n < 0 -> "$_"^(string_of_int (-n))
  | n            -> "$" ^(string_of_int   n )

let expand_neighbor_in_body : Ast.tast -> Ast.tast =
  let open Ast in let open Ikra_type in
      let rec expand (exp,t) = (match exp with
        | Call((Local_var v,Tneighbor(_,m,_)),"[]",exps,None)
            when m = List.length exps
            &&   List.for_all (function 
              |    Literal(Integer_literal _),_ -> true
              |    _                            -> false) exps
            -> let indices = String.concat ""
                 (List.map (function 
                   |    Literal(Integer_literal n),_ -> (ident_of_int n)
                   |    _ -> invalid_arg "never") exps) in
               (Local_var (v ^ indices))

        | Call(exp1,mname,exps,expb) ->

          Call(expand exp1,mname,List.map expand exps,
               Util.Option.maybe expand expb)

        | Literal _ | Constant _ | Local_var _ -> exp
        | Constant_assign(v,exp) -> Constant_assign(v, expand exp)

        | Local_assign(v,exp) -> Local_assign(v,expand exp)
        | Block(exps) -> Block(List.map expand exps)
        | Proc(params,exp) -> Proc(params, expand exp)
        | While(exp1,exp2) -> While(expand exp1,expand exp2)
        | If(exp1,exp2,exp3) -> If(expand exp1,expand exp2,
                                   Util.Option.maybe expand exp3) 
        | For(var,exp1,exp2) -> For(var,expand exp1,expand exp2)
        | Return(exp1) -> Return(Util.Option.maybe expand exp1)
        | Type_decl(decl, exps) -> Type_decl(decl, List.map expand exps)

        | Method_def _ | Yield _ | Array _
          -> invalid_arg "not supported"
      ), t in
      expand 

(* TODO: this must generate functions with the device modifier *)
let body_template : env -> kind -> Cuda.fundef list = fun e kind ->
  let open Cuda_constructors in 
      [__device__inline e.body_ctype e.body_fun_name e.body_args
        (fun _ -> e.element_statements);]

let generate_env : kernel -> env = fun kernel ->

  let fvspecs : (Ast.tast * string) list =
    List.map (fun (fv,i) -> (fv,"fv"^(string_of_int i)))
      (List.combine kernel.free_vars
         (zero_to (List.length kernel.free_vars))) in
  let input_array_itype, input_array_rank, output_array_itype =
    (* type of the array given as an input, i.e., the reciever object *)
    (* NOTE: for initializer, it gives an invalid (i.e., non-array) type *)
    match kernel.kind, kernel.receiver, kernel.return_type with
      | Initialize, _, (Ikra_type.Tarray(rank,_) as typ)
	->	(typ, rank, typ (* dummy *) )
      | Reduce,         (_, (Ikra_type.Tarray(rank,_) as typ)), _
      | UnitlessReduce, (_, (Ikra_type.Tarray(rank,_) as typ)), _
	->	(typ, rank, typ (* dummy *) )
      | Map,    (_, (Ikra_type.Tarray(rank,_) as typ)), _
	->	(typ, rank, Ikra_type.Tarray(rank,kernel.body_type))
      | _ -> invalid_arg "unexpected input types" in

  let sizes =
    List.map (fun dim -> (Ikra_type.Tint, Printf.sprintf "_size%d" dim))
      (zero_to input_array_rank) in

  let block_parameters = match kernel.params with
    |Ast.Array_pattern vps -> vps
    |Ast.Var_pattern _ as vp -> [vp] in


  let expand_neighbor_arg : string -> Ikra_type.t -> (int * int) -> 
    (Cuda.cuda_type * string) list = 
    (* to expand a neighbor argument to a body function to a
       list of neighbor elements.
       expand_neighbor_arg (9,2) "v" Tint
       = [CUDAint, "v$_1$_1";CUDAint, "v$_1$0";CUDAint, "v$_1$1";
       CUDAint,  "v$0$_1";CUDAint,  "v$0$0";CUDAint,  "v$0$1";
       CUDAint,  "v$1$_1";CUDAint,  "v$1$0";CUDAint,  "v$1$1";]
       where names like v$_1$_1 corresponds to the element accessed
       by v[-1,-1] (read _ as minus).
    *)
    fun v ityp ->
      let ctyp = Cuda.cuda_type_of_ikra_type ityp in function
        | (3,1) ->
          map_range (-1) 1 (fun n -> (ctyp, v ^ (ident_of_int n)))
        | (9,2) ->
          List.flatten
          (map_range (-1) 1
            (fun n ->
              map_range (-1) 1
                (fun m ->
                  (ctyp, v ^ (ident_of_int n) ^ (ident_of_int m)))))
        | (27,3) ->
          List.flatten
            (map_range (-1) 1
               (fun l ->
                 List.flatten
                   (map_range (-1) 1
                      (fun n ->
                        map_range (-1) 1
                          (fun m ->
                            (ctyp, v ^ (ident_of_int l)
                              ^ (ident_of_int n) ^ (ident_of_int m)))))))
        | (m,n) -> invalid_arg (Printf.sprintf "unsupported %d-neigbor\
  for %d-dimension array" m n) in

  let rec expand_block_args = fun pats ->
    List.flatten
      (List.map (function
        |Ast.Var_pattern
            (v,Ikra_type.Tneighbor(m,n,t))
          (* Ikra_type.Tarray (n,Ikra_type.Tneighbor(m,n',t))) *)
          (* when n=n' *)
          -> expand_neighbor_arg v t (m,n)
        |Ast.Var_pattern (v,typ) -> [Cuda.cuda_type_of_ikra_type typ, v]
        |Ast.Array_pattern pats -> expand_block_args pats)
      pats) in

          {

            argspecs = List.map make_var 
              (match kernel.kind with
                | Initialize -> (kernel.return_type, "_array")::sizes
                | Reduce ->     (input_array_itype, "_array")
		  ::(element_type_of input_array_itype, "_unit")
		  ::sizes
                | UnitlessReduce -> (input_array_itype, "_array")::sizes
                | Map ->        (input_array_itype, "_array")
		  ::(output_array_itype, "_array2")
		  ::sizes);
            fvspecs;
            kernel_fun_name = "__ikra_kernel_" ^ kernel.name;
            wrapper_name = "__ikra_wrapper_" ^kernel.name;
            body_fun_name = "__ikra_body_" ^kernel.name;
            body_ctype = Cuda.cuda_type_of_ikra_type kernel.body_type;
    (* input_array_itype; *)
    (* output_array_itype; *)
    (* return_itype = kernel.return_type; *)
    (* input_element_type =(match kernel.kind with *)
    (*     |Initialize -> Cuda.CUDAother "INVALID" *)
    (*     | _ -> Cuda.cuda_type_of_ikra_type (element_type_of input_array_itype)); *)
            body_args =
              ((expand_block_args block_parameters)
               @ (List.map (fun ((_,t) as fvp,_) ->
                 try (Cuda.cuda_type_of_ikra_type t,fv_name fvp)
                 with Invalid_argument m -> invalid_arg(m^ " for "^(fv_name fvp))
               ) fvspecs ));
            
            element_statements =
              let type_decls_but_block_parameters () =
                let block_params = Ast.variables_in_pattern kernel.params in
                List.filter (fun (var,_) -> not(List.mem var block_params))
                  kernel.type_decls in
      let open Cuda_constructors in [
        return(
          (Cuda.exp_of_ruby
             (Ast.Type_decl(type_decls_but_block_parameters (),
                            List.map expand_neighbor_in_body kernel.body),
              kernel.body_type)))];
  }

let generate_functions : kernel -> Cuda.fundef list =
  fun kernel ->
    let env = generate_env kernel in
    List.concat
      (List.map (fun template -> template env kernel.kind)
         [body_template; loop_template; wrapper_template])


let generate_preamble : string -> kernel list -> string =
  fun module_name kernels ->
    Printf.sprintf "\
// This file is generated for %s.
#include <ikra.h>
"
      module_name

module Gather_auxiliary_methods = struct

  let predefined_method : Ast.tast -> string -> bool =
    (* to check if a call on a given recivier and mname is a
       predefined one.  For example, n..m is represented as a call
       "self.__PSEUDO_dot2(n,m)", which is not needed to be gathered
       for kernel generation. *)

    fun recevier mname ->
      List.exists (fun (r, m) -> (r recevier) && (m mname))
        [
          (Ast.is_self, fun m -> List.mem m ["__PSEUDO_dot2";"&&";"||";"not@";
                                            "rand"]);
          (fun r -> not(Ast.is_self r)), (fun m -> true);
        ] 

  let rec called_methods : Ast.tast -> string list =
    fun (exp,typ) -> let open Ast in match exp with


      (* TODO: exclude names that are translated in a special way like
         __PSEUDO_dot2.  In future, such a special calls should be
         treated in a different way. *)
      | Call(receiver,mname,args,block) ->
        (if not (predefined_method receiver mname)
          (* Ast.is_self receiver *) then [mname] else [])
        @ called_methods receiver
        @ map_called_methods args
        @ (maybe_called_methods block)

      | Constant_assign(cname,exp1) when Cuda.is_cxx_defined_constant cname
          -> invalid_arg (Printf.sprintf "%s is a constant that supposed to \
 be predefined in the CXX namespace, but is assigned to %s"
                            cname (Ast.string_of exp1))
      | Constant_assign(cname,exp1) -> cname::(called_methods exp1)

      | Constant cname when Cuda.is_cxx_defined_constant cname -> [] 
      | Constant cname                                         -> [cname]

      | Literal _ | Local_var _ -> []

      | Local_assign(_,exp1)
      | Proc(_,exp1)
        -> called_methods exp1
      | While(exp1,exp2) | For(_,exp1,exp2)
        -> (called_methods exp1)@(called_methods exp2)
      | If(exp1,exp2,exp3) ->
        (called_methods exp1)@(called_methods exp2)@(maybe_called_methods exp3)
      | Return(exp1) -> (maybe_called_methods exp1)

      | Block(exps) | Array exps | Type_decl(_, exps)
        -> map_called_methods exps

      | Method_def _ | Yield _ -> invalid_arg "not supported"

    and maybe_called_methods = function
      | Some e -> called_methods e
      | None -> [] 
    and map_called_methods exps =
                       List.concat(List.map called_methods exps) 
                     
    let directly_called {body} =
      List.concat (List.map called_methods body) 

    module MnameSet = Set.Make(String)

    let gather : kernel list -> (string * Ast.tast) list =
      fun kernels ->
        let decls = top_decls kernels in
        let lookup_body mname =
          List.find (fun decl ->
            (Ast.is_method_def mname decl) || (Ast.is_constant_def mname decl))
            decls in
        let rec loop collected = function
          | [] -> collected
          | mname::queue ->
            if List.mem_assoc mname collected 
            then loop collected queue
            else
              let def = lookup_body mname in
            let called = match def with
              | Ast.Method_def _,_ ->
                called_methods (Ast.method_def_body def)
              | Ast.Constant_assign(_,body), _ ->
                called_methods body
              | _ -> invalid_arg "unexpected type of node" in
              loop ((mname,def)::collected) (queue @ called) in
        loop [] (List.concat (List.map directly_called kernels))
end

let generate_aux_functions : string -> kernel list -> string =
  (* to generate a string representation of C functions that are
     called from kernels *)

  (* TODO: those functions must be device function (or global),
     which can be called from kernels *)

  fun module_name kernels ->
    let aux_funs string_of =
      List.map
        (fun (_,def) -> string_of(
          Cuda.prepend_modifier "__device__"
            (Cuda.decl_of_ruby_ast def)))
        (Gather_auxiliary_methods.gather kernels) in
    String.concat "\n"
      ((aux_funs 
          (fun d -> (Cuda.prototype_of_decl d) ^ ";"))
       @ (aux_funs Cuda.string_of_decl))
      
let wrapper_registration : Cuda.exp -> string -> Cuda.fundef -> Cuda.statement =
  let dec x = x - 1 in
  let open Cuda_constructors in fun cKernel kernel_name -> function
    | {Cuda.formal_params; Cuda.name} ->
      _exp(
        rb_define_module_function
          (cKernel, _str kernel_name,
           _cast _VALUE_func_t (Cuda.Var name),
           _num (dec (List.length formal_params))))

let generate_initializer : string -> kernel list -> Cuda.fundef list -> string =
  fun module_name kernels wrappers ->
    let init_module_name = "Init_" ^ module_name  in
    let module_name = "Ikra_Kernels__" ^ module_name  in
    let fundef =
      let open Cuda_constructors in
          extern_C void init_module_name [] (fun _ ->
            _let _VALUE "cKernel"
              (rb_define_module(_str module_name))
              (fun cKernel -> 
                ((List.map2 (fun kernel -> wrapper_registration cKernel kernel.name)
                      kernels wrappers) @
                   [return_]))) in
      Cuda.string_of_decl (Cuda.Function_decl fundef)

let create_makefile_opt = "\\$CUDA=true;"

let generate_makefile : generate_only:bool -> Convert_env.t -> unit =
  (* (generate_makefile ~generate_only dir module_name) prepaers a
     Makefile for compiling generated files as an Ruby extension.
     Unless generate_only is specified, the make is genareted and
     executed. *)

  fun ~generate_only gen_env ->
    Util.with_cd (Filename.dirname
                    (Convert_env.get_cuda_file_name gen_env))
      (fun _ ->
        (fun p -> if generate_only then
            (let ochan =
               open_out (Convert_env.get_build_script_name gen_env) in
             p (fun args -> 
               output_string ochan (String.concat " " ("ruby"::args));
               output_string ochan "\n";
               "written.")
               (fun args ->
                 output_string ochan (String.concat " " args);
                 output_string ochan "\n";
                 "written");
             close_out ochan)
             else p (Util.run_ruby ~redirect:true)
            (Util.run_command ~redirect:true))
          (fun run_ruby run_command ->

            let require_name = Convert_env.get_require_name gen_env in
            let _ = (run_ruby ["-e"; 
                               "\"" ^ create_makefile_opt ^ 
                               "['.',ENV['HOME']].find do |d| p=File.expand_path('.ikra.conf',d); File.exists?(p) && load(p); end; PARRAY_C_DIR=\\\"#{IKRA_HOME}/proto/parray\\\"; load \\\"#{IKRA_HOME}/src/extconf.rb\\\"; \
                                   create_makefile '"^require_name^"'\"" ]) in
      (* TODO make the make execution silent as long as it runs
         successfully.  Make a general command invocation function in
         util.ml. *)
            let _ = try run_command ["make"; "1>&2"]
              with Util.External_command_error(command,stdout,stderr) as e ->
                Printf.printf "Compilation error:\n%s\nOutput:\n%s\n\
 Error:\n%s\nSource code:\n%s\n"
                  command stdout stderr (Util.read_lines gen_env.Convert_env.cuda_file_name); 
                raise e in
            ()))

 let rec last = function
  | [x] -> x
  | _::tl -> last tl
  | [] -> invalid_arg "empty list"

let cuda_code_of : string -> kernel list -> string =
  (* (cuda_code_of module_name kernels) constructs a CUDA program from
     the given kernels under the given module_name. *)

  fun module_name kernels ->
    let kernel_funs = List.map generate_functions kernels in
    let wrappers = List.map last kernel_funs in
    (* let wrappers = List.map (function [_;_;wrapper] -> wrapper *)
    (*   | _ -> invalid_arg "not three kernels") kernel_funs in *)
    (String.concat "\n"
       ([generate_preamble module_name kernels]
        @ [generate_aux_functions module_name kernels]
        @ (List.map 
             (fun d ->Cuda.string_of_decl (Cuda.Function_decl d))
             (List.flatten kernel_funs))
        @ [generate_initializer module_name kernels wrappers]))

(* for testing *)
let generate_tuple_type_decls = fun _ -> invalid_arg "not yet"
