(* constraint resolution *)

type craint_t = Type_constraints.Constr.t
type craints_t = Type_constraints.Constraints.t
type subst_t = (Ikra_type.t * Ikra_type.t) list

let empty_subst : subst_t = [] 


let rec apply_to_type : subst_t -> Ikra_type.t -> Ikra_type.t =
  let open Ikra_type in 
  fun subst -> function
    | Tarray(dim,base) -> Tarray(dim,apply_to_type subst base) 
    | Tvar(n) as t -> (try List.assoc t subst with Not_found -> t)
    | Ttuple(typs) -> Ttuple(List.map (apply_to_type subst) typs)
    | Tneighbor(m,n,t) -> Tneighbor(m,n, apply_to_type subst t)
    | Tproc(args,ret,block) ->
      Tproc(List.map (apply_to_type subst) args,
            apply_to_type subst ret,
            Util.Option.maybe (apply_to_type subst) block)
    | t -> t

let apply_to_subst : subst_t -> subst_t -> subst_t =
  fun subst ->
    List.map (fun (t1,t2) -> (t1, apply_to_type subst t2))

let apply_to_craint : subst_t -> craint_t -> craint_t =
  let open Type_constraints.Constr in
  fun subst ->
      let app typ = apply_to_type subst typ in function
    | Less_than(t1,t2)        -> Less_than(app t1, app t2)
    | Equal(t1,t2)            -> equal(app t1, app t2)
    | Has_method(tr,m,tp)     -> Has_method(app tr,m,app tp)
    | Defines_method(tr,m,tp) -> Defines_method(app tr,m,app tp)

let apply_to_craints : subst_t -> craints_t -> craints_t =
  let map f set = Type_constraints.Constraints.fold
    (fun elt s -> Type_constraints.Constraints.add (f elt) s)
    set Type_constraints.Constraints.empty in
  fun subst -> map (apply_to_craint subst) 

let rec apply_to_parameter_pattern
    : subst_t -> Ast.parameter_pattern -> Ast.parameter_pattern =
  let open Ast in
  fun subst -> function 
      | Var_pattern(var, typ) -> Var_pattern(var, apply_to_type subst typ)
      | Array_pattern(patterns) ->
	Array_pattern(List.map (apply_to_parameter_pattern subst)  patterns)

let apply_to_ast : subst_t -> Ast.tast -> Ast.tast = fun subst ->
  let apply_to_type_decl (var,typ) = (var, apply_to_type subst typ) in
  let rec apply (term,typ) =
    let typ = apply_to_type subst typ in
    let term = let open Ast in match term with 
      | Call(receiver, mname, operands, block)
	-> Call(apply receiver, mname, List.map apply operands,
		Util.Option.maybe apply block)
      | Literal _ 
      | Constant _ 
      | Local_var _ -> term
      | Constant_assign(var,rhs) -> Constant_assign(var,apply rhs)
      | Local_assign(var,rhs) -> Local_assign(var,apply rhs)
      | Block(exps) -> Block(List.map apply exps)
      | Proc(params,exp)
	-> Proc(apply_to_parameter_pattern subst params, apply exp)
      | Method_def(mname, params, body)
	-> Method_def(mname, apply_to_parameter_pattern subst params,
		      apply body) 
      | Yield(exp) -> Yield(apply exp)
      | Array(exps) -> Array(List.map apply exps)
      | While(cond,body) -> While(apply cond,apply body)
      | If(cond,then_exp,else_exp)
        -> If(apply cond,apply then_exp,Util.Option.maybe apply else_exp)
      | For(var,range,body) -> For(var,apply range,apply body)
      | Return(exp1) -> Return(Util.Option.maybe apply exp1)

      | Type_decl(tdecls, exps)
        -> Type_decl(List.map apply_to_type_decl tdecls, List.map apply exps) in
    (term,typ) in
  apply

let process_rule : craint_t -> (craints_t * subst_t)
  -> (craints_t * subst_t) =
  let open Type_constraints.Constraints in
  let open Type_constraints.Constr in
  let open Ikra_type in
      fun c (craints_processed, subst) -> match c with
	(* the following two rules---removal of reflexive
	   constraints---should come earlier than the the other equality
	   rules; otherwise, it will generate a substitution like t0->t0. *)
	| Less_than(t1,t2) | Equal (t1,t2) when t1=t2 ->
	  (craints_processed, subst)
	| Equal(Tvar n,typ) | Equal(typ,Tvar n) ->
	  let new_subst = [Tvar n,typ] in
	  (apply_to_craints new_subst craints_processed,
	   new_subst @ (apply_to_subst new_subst subst))
	| Equal(Tarray(d1,t1),Tarray(d2,t2)) when d1=d2 ->
	  (add (equal(t1,t2)) craints_processed, subst)
	| rel -> (add rel craints_processed, subst) 
	  
let unify :(craints_t * subst_t) -> (craints_t * subst_t) =

  fun (craints,subst) -> 
    Type_constraints.Constraints.fold
      (fun c (craints,subst') ->
	process_rule (apply_to_craint subst' c) (craints,subst'))
      craints
      (Type_constraints.Constraints.empty, subst)

let remove_cycle :(craints_t * subst_t) -> (craints_t * subst_t) =

  (* to generate equality constraints from chains of subtype relations
     that make cycles in a given set of constraints

     remove_cycle([ t0<:t1; t1<:t0 ],s)
     -> ([ t0<:t1; t1<:t0; t1=t0 ],s)
  *)


  let extract_subtype_relations
      : craints_t -> (Ikra_type.t * Ikra_type.t) list =
    fun craints ->
      List.map (function
	| Type_constraints.Constr.Less_than(t1,t2) -> (t1,t2)
	| _ -> invalid_arg "never") 
	(List.filter (function
	  | Type_constraints.Constr.Less_than _ -> true
	  | _ -> false)
	   (Type_constraints.Constraints.elements craints)) in
  let extract_types : (Ikra_type.t * Ikra_type.t) list -> Ikra_type.t list =
    fun l ->
      List.fold_left (fun set elem ->
	if List.mem elem set then set else elem::set)
	[]
	((List.map (fun (x,_) -> x) l) @
	    (List.map (fun (_,x) -> x) l)) in
  let generate_equalities : Ikra_type.t list list -> craints_t -> craints_t =
    fun sccs craints ->
      List.fold_left (fun craints scc ->
	let hd = List.hd scc in
	List.fold_left (fun craints t ->
	  Type_constraints.Constraints.add 
	    (Type_constraints.Constr.equal(hd,t))
	    craints) craints (List.tl scc))
	craints sccs in

  fun (craints, subst) ->
    let subtype_relations = extract_subtype_relations craints in
    let elements = extract_types subtype_relations in
    let sccs = Tarjan.sccs elements subtype_relations in
    let craints' = generate_equalities sccs craints in
    (craints', subst)

let shortcut : (craints_t * subst_t) -> (craints_t * subst_t) =

  (* to generate inequality constraints from pairs of transitive
     subtype relations

     shortcut([ t0<:t1; t1<:t2 ],s) -> ([ t0<:t1; t1<:t2; t0<:t2 ],s)
  *)

  let extract_subtype_relations
      : craints_t -> (Ikra_type.t * Ikra_type.t) list =
    fun craints ->
      List.map (function
	| Type_constraints.Constr.Less_than(t1,t2) -> (t1,t2)
	| _ -> invalid_arg "never") 
	(List.filter (function
	  | Type_constraints.Constr.Less_than _ -> true
	  | _ -> false)
	   (Type_constraints.Constraints.elements craints)) in

  let gather_by_first
      : (Ikra_type.t * Ikra_type.t) list
      -> (Ikra_type.t * (Ikra_type.t list)) list =
    fun pairs ->
      let sorted = List.sort
        (fun (t1,_) (t2,_) -> Type_constraints.Constr.compare_types t1 t2)
        pairs in
      let rec gather = function
        | (None,                    [])        -> []
        | (None,                    (k,v)::tl) -> gather (Some(k,[v]), tl)
        | (Some(prev_k, prev_vals), [])        -> [prev_k, prev_vals]
        | (Some(prev_k, prev_vals), (k,v)::tl) when prev_k = k 
          -> gather (Some(prev_k, v::prev_vals), tl)
        | (Some(prev_k, prev_vals), (k,v)::tl) 
          -> (prev_k, prev_vals) :: (gather (Some(k, [v]), tl)) in
      gather (None, sorted) in
  
  let rec merge
      : ((Ikra_type.t*'a) list * (Ikra_type.t*'b) list)
      -> (Ikra_type.t * 'a * 'b) list = function
        | ([],         [])          -> []
        | ((k1,v1)::t1,[])          -> (k1,v1,[])::(merge (t1,[]))
        | ([],         (k2,v2)::t2) -> (k2,[],v2)::(merge ([],t2))
        | ((k1,v1)::t1,(k2,v2)::t2) when k1=k2 -> (k1,v1,v2)::(merge (t1,t2))
        | ((k1,v1)::t1,(k2,v2)::t2)
            when Type_constraints.Constr.compare_types k1 k2 < 0 ->
          (k1,v1,[])::(merge (t1,(k2,v2)::t2))
        | ((k1,v1)::t1,(k2,v2)::t2)
            (* when Type_constraints.Constr.compare_types k1 k2 < 0 *) -> 
          (k2,v2,[])::(merge ((k1,v1)::t1,t2)) in
  
  let build_table
      : (Ikra_type.t * Ikra_type.t) list
      -> (Ikra_type.t * (Ikra_type.t list) * (Ikra_type.t list)) list =
    fun inequalities ->
      let greaters = gather_by_first inequalities in
      let swap (a,b) = (b,a) in
      let lesses   = gather_by_first (List.map swap inequalities) in
      merge (lesses, greaters) in

  let make_pairs 
      : (Ikra_type.t list) ->  (Ikra_type.t list)
    -> Type_constraints.Constr.t list =
    fun lesses greaters ->
      List.flatten
        (List.map (fun l ->
          List.map (fun g -> Type_constraints.Constr.Less_than(l,g))
            greaters) lesses) in

  let scan_table
      : (Ikra_type.t * (Ikra_type.t list) * (Ikra_type.t list)) list 
      -> craints_t -> craints_t =
    List.fold_right (fun (_, lesses, greaters) ->
      List.fold_right Type_constraints.Constraints.add 
        (make_pairs lesses greaters)) in

  fun (craints, subst) ->
    (* let _ = print_string "make shortcuts\n" in *)
    let subtype_relations = extract_subtype_relations craints in
    let table = build_table subtype_relations in

    (* let _ = *)
    (*   let s = Ikra_type.string_of in *)
    (*   let ss types = "{" ^ (String.concat "," (List.map s types)) ^ "}" in *)
    (*   List.iter (fun (k, l, g) -> *)
    (*     Printf.printf "%s < %s < %s\n" (ss l) (s k) (ss g)) table in *)

    let craints' = scan_table table craints in
    (craints', subst)
    
let set_print = fun s -> "{" ^ 
  (String.concat "," (List.map Type_constraints.Constr.string_of
                        (Type_constraints.Constraints.elements s))) ^ 
  "}"

let trace : ((craints_t * subst_t) -> unit) ref =
  ref (fun _ -> ())
  (* ref (fun (craints,_) ->  *)
  (*   print_string (set_print craints); *)
  (*   print_string "\n") *)

let resolve_one : (craints_t * subst_t) -> (craints_t * subst_t) =
  fun state ->
    (* print_string "resolve_one\n"; *)
    shortcut (remove_cycle (unify state))

(* The second level constraint solving

   At this level, we use domain specific knowledge such as

   - the language has not subtyping except for Int <: Float 
   and other few exceptions;

   - the language prefers smaller types (when we can use both Int and
   Float, we should use Int);

   - arithmetics like Int * Int -> Int; and arrays.

   Those knowledge should be used after we have done the first level
   constraint solving.  After applied the second level, the first
   level should be done again.

*)

let arith_ops = ["+";"-";"*";"%";"/";"**";"<<";">>";"|";"&"]
let compare_ops = ["=="; "<"; ">"; "<="; ">="; "!="]
let boolean_ops = ["&&";"||"]

let neighbor_alist : (string * (int * int)) list =
  (* a list of neighbor method specifications given as a tuple of the
     method name, the number of points, and the array rank.  For
     example, ("neighbor9", 9, 2) corresponds to the 9-neighbors
     created by a method "neighbor9" from a two dimensional array.  *)
  List.map (fun (points,rank) ->
    ("neighbor"^(string_of_int points), (points, rank))) 
    [
      3,1;
      9,2;
     27,3;
    ]

let neighborE_alist : (string * (int * int)) list =
  (* a list of neighbor method specifications given as a tuple of the
     method name, the number of points, and the array rank.  For
     example, ("neighbor9", 9, 2) corresponds to the 9-neighbors
     created by a method "neighbor9" from a two dimensional array.  *)
  List.map (fun (points,rank) ->
    ("neighborE"^(string_of_int points), (points, rank))) 
    [
      3,1;
      9,2;
     27,3;
    ]


type side_condition = craint_t -> bool

let show_trace = ref false            (* TODO: move this variable to the top *)

(* TODO: make sure that every rule application says somthing *)
let replace : craint_t -> string -> craint_t list  -> craint_t list =
  let open Type_constraints.Constr in
  fun old name news ->
    if !show_trace then
      Printf.printf "rule %s:\n\t%s -> %s\n"
        name (string_of old) (String.concat ";" (List.map string_of news))
    else ();
    news 

(* TODO: reorganize the rules; they are redundant; Ikra specific rules
   should be externalized; ... *)

let second_level_rules : craint_t
    -> (side_condition -> craint_t option) -> craint_t list =
  let open Type_constraints.Constr in
  let open Ikra_type in fun craint side_matcher ->
    let replace = replace craint in
    let is_number_type : Ikra_type.t -> bool = function
      | Tint | Tfloat -> true
      | a ->
        Util.Option.is_some (side_matcher (function
          | Less_than(Tint, a') | Less_than(a',Tfloat) -> a=a'
          | _                                          -> false)) in
    let is_boolean_type : Ikra_type.t -> bool = function
      | Tbool -> true
      | a ->
        Util.Option.is_some (side_matcher (function
          | Less_than(Tbool, a') | Less_than(a',Tbool) -> a=a'
          | _                                          -> false)) in

    match craint with
  (* t1 <: Int => t1 = Int *)
  | Less_than(Tvar v,Tint) -> replace "Int is least"
    [equal(Tvar v,Tint)]
  (* t1 <: Bool => t1 = Bool *)
  | Less_than(Tvar v,Tbool) -> replace "Bool is least"
    [equal(Tvar v,Tbool)]

  (* The next rule is not safe --- moved to the third level *)
  (* (\* Float <: t1 => t1 = Float *\) *)
  (* | Less_than(Tfloat,Tvar v) -> [equal(Tvar v,Tfloat)] *)

  (* Tuple(t0..tN) <: Tuple(s0..sN) => ti = si *)
  (* NOTE: no covariance at this point... *)
  | Less_than(Ttuple(ts),Ttuple(ss)) 
  | Equal    (Ttuple(ts),Ttuple(ss)) when List.length ts = List.length ss
    -> replace "structural for tuples"
    (List.map2 (fun t s -> equal(t,s)) ts ss)

  (* Neighbor(m,n,t) <: Tuple(m,n,t') => ti = si *)
  | Less_than(Tneighbor(m,n,t),Tneighbor(m',n',t'))
      when m=m' && n=n'  -> replace "structural for neighbor"
    [equal(t,t')]


  (* t1*..*tN -> tR <: s1*..*sN -> sR => si <: ti && tR <: sR *)
  | Less_than(Tproc(args,ret,Some block), Tproc(args',ret',Some block'))
      when List.length args = List.length args' -> replace "proc/block sub"
    ((Less_than(ret,ret'))
    :: (Less_than(block',block))
    :: (List.map2 (fun t t' -> Less_than(t',t)) args args'))
  | Less_than(Tproc(args,ret,None), Tproc(args',ret',None))
      when List.length args = List.length args' -> replace "proc sub"
    ((Less_than(ret,ret'))
     :: (List.map2 (fun t t' -> Less_than(t',t)) args args'))
  | Equal(Tproc(args,ret,Some block), Tproc(args',ret',Some block'))
      when List.length args = List.length args' -> replace "proc/block equal"
    ((equal(ret,ret'))
    :: (equal(block',block))
    :: (List.map2 (fun t t' -> equal(t',t)) args args'))
  | Equal(Tproc(args,ret,None), Tproc(args',ret',None))
      when List.length args = List.length args' -> replace "proc equal"
    ((equal(ret,ret'))
    :: (List.map2 (fun t t' -> equal(t',t)) args args'))

  | Equal(Tproc([Ttuple(args)],ret,block), t2) 
  | Equal(t2, Tproc([Ttuple(args)],ret,block)) -> replace "proc(tuple) equal"
    [equal(Tproc(args,ret,block), t2)]
      
  (* Object <: t1 => t1 = Object *)
  | Less_than(Tobject, t) -> replace "object is top"
    [equal(t,Tobject)]

  (* Int <: Float => none *)
  | Less_than(Tint,Tfloat) -> replace "eliminate int<float" []
  (* t1 <: Object => none *)
  | Less_than(_,Tobject) -> replace "eliminate any<object" []

  (* Int has m : 'a -> 'b => 'a <: Float and 'a = 'b *)
  (* | Has_method(Tint, op, Tproc([a], b, None)) when List.mem op arith_ops *)
  (*   -> [Less_than(a,Tfloat); equal(a,b)] *)

  (* Float has m : 'a -> 'b => 'a <: Float and Float = 'b *)
  (* | Has_method(Tfloat, op, Tproc([a], b, None)) when List.mem op arith_ops *)
  (*   -> [Less_than(a,Tfloat); equal(Tfloat,b)] *)

  (* the above two rules are subsumed by the next rule.  delete when
     confident *)

  (* 
                    'a @ 'b -> 'c && 'a is number
                  ----------------------------------
                  Int <: 'b  && 'a <: 'c && 'b <: 'c
  *)
  | Has_method(a, op, Tproc([b], c, None)) 
      when List.mem op arith_ops
	&& is_number_type a (* Util.Option.is_some (side_matcher (function *)
   (* | Less_than(Tint, a') | Less_than(a',Tfloat) -> a=a' | _ -> false)) *)
    -> replace "arithmetic" 
        [(* craint;  *)Less_than(Tint,b); Less_than(a,c); Less_than(b,c) ]

  (* boolean conjuctions and disjunctions *)
   | Has_method(a, op, Tproc([b], c, None))  
       when List.mem op boolean_ops 
         && is_boolean_type a (* Util.Option.is_some (side_matcher (function *)
     -> (* invalid_arg "boolean!!!\n";  *)replace "boolean"  
         [Less_than(Tbool,b); Less_than(a,c); Less_than(b,c) ] 

  (* boolean negation *)
  | Has_method(dummy, "not@", Tproc([a], c, None)) 
      when is_boolean_type a (* Util.Option.is_some (side_matcher (function *)
    -> replace "boolean" 
        [Less_than(a,c) ]

  (* 'a has "==" : 'b -> 'c => 'c = Bool *)
  | Has_method(a, op, Tproc([b], c, None)) when List.mem op compare_ops
    -> replace "equal op." [equal(c,Tbool)] (* can we delete the original rule? *)

  (* Int has "-@" : () -> 'c => 'c = Int *)
  (* Float has "-@" : () -> 'c => 'c = Float *)
  | Has_method(a, "-@", Tproc([], c, None)) -> replace "unary minus"
    [equal(c,a)]

  | Has_method(a, "abs", Tproc([], b, None))
    when a = Tint || a = Tfloat ||
     Util.Option.is_some (side_matcher (function
        | Less_than(Tint, a') | Less_than(a', Tfloat) -> a=a'
        | _ -> false))
      -> replace "abs" [equal(b,a)]

  | Has_method(a, "floor", Tproc([], b, None))
    when a = Tint || a = Tfloat ||
      Util.Option.is_some (side_matcher (function
        | Less_than(Tint, a') | Less_than(a', Tfloat) -> a=a'
        | _ -> false))
      -> replace "floor" [equal(b,Tint)]

  (* 'a has type_assert : 'b -> 'c => 'a = 'c *)
  | Has_method((Tvar _) as a, "type_assert", Tproc([_],      c, None)) 
  | Has_method(a,             "type_assert", Tproc([Tvar _], c, None))
      as craint -> replace "type_assert propagate1"
    [equal(a,c); craint]		(* if one of two types are still
					   a type vaiable, keep it *)
  | Has_method(a, "type_assert", Tproc([b], c, None)) as craint -> 
    if a = b
     then ((* Printf.printf "type assertion succeeded: %s = %s\n" *)
	   (*   (Ikra_type.string_of a) (Ikra_type.string_of b); *)
           replace "type_assert success"
	   [equal(a,c)])
     else ((* Printf.printf "type assertion failed: %s != %s\n" *)
	   (*   (Ikra_type.string_of a) (Ikra_type.string_of b); *)
           replace "type_assert propagate2"
	   [craint; equal(a,c)])

  (* TODO: rewrite the rules to cope with a constraint whose block
     type is a type variable. *)

  (* rules for arrays *)
  (* Class<Array> has new : 'a -> &('b -> 'c) -> 'd *)
  | Has_method(Tclass("Array"), "new", Tproc([tSize], tArray, Some(Tproc([tBarg],tBret, None))))
    -> replace "Array.new 1d"
    [equal(tSize,Tint); equal(tBarg,Tint); equal(tArray, Tarray(1, tBret))]

  (* NOTE: we use new_ rather than new.  Array.new(size,init) is in
     the original Ruby API and creates one-dimension array of size. *)
  (* Class<Array> has new_ : 'a1*'a2*...*'aN -> &('b1*'b2*...*'bN -> 'c) -> 'd
     => 'd = Array(N,'d) *)
  | Has_method(Tclass("Array"), "new_", Tproc(tSizes, tArray, Some(tBlock(* Tproc(tBargs,tBret,None) *))))
      (* when List.length tSizes = List.length tBargs *)
      -> let rank = List.length tSizes in
         let tBret = Typeinf.new_type_var () in
         replace "Array.new multi-dimensional"
	   ((equal(tArray, Tarray(rank, tBret)))
         ::(Less_than(tBlock,
                      Tproc(List.map (fun _ -> Tint) tSizes, tBret, None)))
	 ::(List.map (fun tSize -> equal(tSize,Tint)) tSizes))
	 (* @ (List.map (fun tBarg -> equal(tBarg,Tint)) tBargs) *)

  (* Array(1,t0) has [] : 'a -> 'b => 'a = Int and 'b = t0 *)
  | Has_method(Tarray(1,t0), "[]", Tproc([t1], t2, None))
    -> replace "aref" [equal(t1,Tint); equal(t0,t2)]

  (* Array(n,Tuple<t1..tm>) has map  : () -> &('b1*..*'bn -> 'c) -> 'd
     => 'bi <: ti and 'd = Array(n,'c) *)
  (* TODO: support nested tuples as this only supports zips of arrays. *)
  | Has_method(Tarray(n,Ttuple(tElms)), "map", Tproc([], tResult, Some(tBlock(* Tproc(tBargs,tBret,None) *))) )
    -> let tBret = Typeinf.new_type_var () in
       replace "map on zipped arrays"
       [ (equal(tResult, Tarray(n,tBret)));
         (equal(tBlock, Tproc(tElms,tBret,None)));
       ]
    (* :: (List.map2 (fun tBarg tElm -> Less_than(tBarg,tElm)) tBargs tElms) *)

  (* Array(n,t0) has map  : () -> &('b -> 'c) -> 'd
     => t1 = 'b and 'd = Array(n,'c) *)
  | Has_method(Tarray(n,tElm), "map", Tproc([], tResult, Some(tBlock(* Tproc([tBarg],tBret,None) *))) )
    -> let tBret = Typeinf.new_type_var () in
       replace "map on simple array"
         [ (* equal(tElm,tBarg); *)
         equal(tResult, Tarray(n,tBret));
         equal(tBlock, Tproc([tElm],tBret,None));
       ]


  (* Array(n,'a) has reduce  : ('b) -> &('c * 'd -> 'e) -> 'f
     => 'a<:'e and  'b<:'e  'c = 'd = 'e = 'f *)
  | Has_method(Tarray(n,tElm), "reduce", Tproc([tUnit], tResult,
	                                       Some(tBlock)) )
    -> replace "reduce with unit"
       [ Less_than(tElm,  tResult);
	 Less_than(tUnit, tResult);
         equal    (tBlock,Tproc([tResult;tResult],tResult,None));
       ]

  (* Array(n,'a) has reduce  : () -> &('c * 'd -> 'e) -> 'f
     => 'a<:'e and  'c = 'd = 'e = 'f *)
  | Has_method(Tarray(n,tElm), "reduce", Tproc([], tResult,
	       Some(tBlock(* Tproc([tBarg1;tBarg2],tBret,None) *)) ))
    -> replace "reduce without unit"
       [ Less_than(tElm,  tResult);
         equal    (tBlock,Tproc([tResult;tResult],tResult,None));
	 (* equal    (tBarg1,tResult); *)
	 (* equal    (tBarg2,tResult); *)
	 (* equal    (tBret, tResult); *) ]

  (* Array(n,'a) has zip : ('b1 * .. 'bn) -> 'c 
     => 'bi = Array(n,'di) && 'c = Array(n,Tuple('a,'d1,..,'dn)) *)
  | Has_method(Tarray(n,tElm), "zip", Tproc(argTypes, tResult, None))
    -> let argElementTvars =
         List.map (fun _ -> Typeinf.new_type_var ()) argTypes in
       replace "zip"
         ((equal(tResult, Tarray(n,Ttuple(tElm::argElementTvars))))
       ::(List.map2 (fun targ telm -> equal(targ, Tarray(n,telm)))
            argTypes argElementTvars))

  (* TODO: neighborE is very ad-hoc; consider unification *)
  | Has_method(Tarray(rank,tElm), mname, Tproc([], tResult, None))
      when List.mem_assoc mname neighborE_alist 
    -> let (num_points, rank') = List.assoc mname neighborE_alist in
       if rank=rank' then
       replace "neighborE"
       [equal(tResult, Tarray(rank, (* TODO: change the tuple type to
                                    the neighbor type because we need
                                    to distinguish zipped arrays and
                                    neighbor arrays *)
                              Tneighbor(num_points,rank,tElm)));]
       else
         (Printf.printf "%s should be applied to %d dimensional arrays,\
 but to %d.\n" mname rank' rank;
          [craint])                         (* keep the constraint *)

  (* Array(1,'a) has neighbor3 : ('b) -> 'c
     => 'c = Array(1,Tuple('a,'a,'a)) && 'b = 'a *)
  (* TODO: support 5, 9, etc.-neighbors *)
  | Has_method(Tarray(rank,tElm), mname, Tproc([tPad], tResult, None))
      when List.mem_assoc mname neighbor_alist 
    -> let (num_points, rank') = List.assoc mname neighbor_alist in
       if rank=rank' then
       replace "neighbor with pad"
       [equal(tResult, Tarray(rank, (* TODO: change the tuple type to
                                    the neighbor type because we need
                                    to distinguish zipped arrays and
                                    neighbor arrays *)
                              Tneighbor(num_points,rank,tElm)));
        equal(tElm, tPad);]
       else
         (Printf.printf "%s should be applied to %d dimensional arrays,\
 but to %d.\n" mname rank' rank;
          [craint])                         (* keep the constraint *)

  (* TODO: support 5, 9, etc.-neighbors *)
  | Has_method(Tarray(rank,tElm), mname, Tproc([], tResult, None))
      when List.mem_assoc mname neighbor_alist 
    -> let (num_points, rank') = List.assoc mname neighbor_alist in
       if rank=rank' then
       replace "neighbor" 
       [equal(tResult, Tarray(rank, (* TODO: change the tuple type to
                                    the neighbor type because we need
                                    to distinguish zipped arrays and
                                    neighbor arrays *)
                              Tneighbor(num_points,rank,tElm)));
       ]
       else
         (Printf.printf "%s should be applied to %d dimensional arrays,\
 but to %d.\n" mname rank' rank;
          [craint])                         (* keep the constraint *)

  (* Neighbor(m,n,'a)  has [] : ('b1*..*'bn) -> 'c
     => 'bi = int && 'c = 'a *)
  (* TODO: use dependent type to constrain the range of numbers *)
  | Has_method(Tneighbor(m,dim,tElm), "[]", Tproc(tIdxs, tResult, None))
      when dim = (List.length tIdxs) 
    -> replace "aref on neighbors"
    ((equal(tElm,tResult))
    :: (List.map (fun tIdx -> equal(Tint, tIdx)) tIdxs))

  (* Math functions *)
  | Has_method(Tclass "Math", mname, Tproc([t0], t1, None) )
      when List.mem mname ["acos";"acosh";"asin";"asinh"; "atan"; "atanh";
                           "cbrt"; "cos"; "cosh"; "erf"; "erfc"; "exp";
                           "gamma"; "lgamma"; "log"; "log10"; "log2";
                           "rsqrt"; "sin"; "sinh"; "sqrt"; "sqrt"; "tan";
                           "tanh";] -> replace "Math functions"
    [ Less_than(t0,Tfloat);
      equal(t1,Tfloat) ]

  (* rand : int -> int | float -> float *)
  | Has_method(Tobject, mname, Tproc([t0], t1, None) )
      when List.mem mname ["rand" ] -> replace "rand"
    [ Less_than(t0,Tfloat); equal(t1,t0) ]

  (* TODO: support other math functions that take two arguments
     "atan2"; "frexp"; "hypot"; "ldexp"; "log", or that return two
     values *)

  (* type conversion *)
  | Has_method(_, mname, Tproc([], t0, None))
      when List.mem_assoc mname ["to_i", Tint; "to_f", Tfloat] ->
    replace "coercion"
    [equal(t0, List.assoc mname ["to_i", Tint; "to_f", Tfloat])]

  (* importing Ruby arrays into Ikra

     e.g., a_Ruby_array.to_a_i() constructs a 1d-PArray of integer

  *)
  | Has_method(_, mname, Tproc([], t0, None))
      when Str.string_match (Str.regexp
                               "to_a_\\([if]\\)\\([0123456789]*\\)") mname 0 ->
    let element_type = (match Str.matched_group 1 mname with
      | "i" -> Tint | "f" -> Tfloat) in
    let rank = match Str.matched_group 2 mname with
      | "" -> 1 | n -> int_of_string n in
    replace "coercion to an array"
      [equal(t0, Tarray(rank, element_type))]

  (* range operator x..y *)
  | Has_method(Tobject, "__PSEUDO_dot2", Tproc([t1; t2], t0, None)) ->
    replace "dot-dot operator"
    [equal(t1, Tint); equal(t2, Tint); equal(t0, Trange)]
      
  (* 'a has m : 'b -> 'c && 'a defines m : 'd -> 'e => 'b = 'd && 'c = 'e *)
  (* NOTE: this is not actually Ikra specific *)
  (* TODO: allow block parameter *)
  (* TODO: make it polymorphic; instead of making equality constraints
     between formal and actual parameters and return values, it will
     make copies of type variables in the defines_method constraint
     (and those generated in its body ???), and then make constraints.
     This is basically the same what let polymorphism does.  (??? This
     will require defines_method to keep the constraints that are
     generated from its body.) *)
  | Has_method(a, op, Tproc(b, c, None)) as craint ->
    (match (side_matcher
	      (function | Defines_method(a',op',Tproc(_,_,None))
                -> a=a' && op = op'
		        | _ -> false)) with
      |Some(Defines_method(_, _, Tproc(b', c', None)))
	  -> replace "call user defined method"
        ((equal(c,c'))::(List.map2 (fun bi bi' -> equal(bi,bi')) b b'))
      | _ -> [craint])

  | c -> [c]

let third_level_rules : craint_t
    -> (side_condition -> craint_t option) -> craint_t list =
  let open Type_constraints.Constr in
      let open Ikra_type in fun craint side_matcher -> match craint with

  (* Float <: t1 => t1 = Float *)
  | Less_than(Tfloat,Tvar v) -> [equal(Tvar v,Tfloat)]

  (* Bool <: t1 => t1 = Bool  ---  this rule will fix a type variable to
     the minimal type among the possibilities *)
  | Less_than(Tbool,Tvar v) ->[equal(Tvar v,Tbool)]

  (* Array(n,t0) <: t1 => t1 = Array(n,t0) --- array types are
     invariant *)
  | Less_than(Tarray _ as typ, Tvar v) -> [equal(Tvar v, typ)]

  (* NOTE: the next rule is now at the third level because at the
     second level, we have a rule to resolve method call/definition
     pairs including those on Object types. *)

  (* TODO: combine the next three rules into one. *)
  (* TODO: handle the cases when block does not have type Tproc. *)
  (* Object has mname : 'a1*..*'an -> &('b1*..*bm -> 'c) -> 'd
     => ai<:Object and bi<:Object and c = Object and d = Object *)
  | Has_method(Tobject, any_method, Tproc(tArgs, tResult, block))
    -> (* print_string ("object invocation:" *)
       (*               ^(Type_constraints.Constr.string_of craint)^"\n"); *)
       equal(tResult,Tobject)
    :: (List.map (fun tArg -> Less_than(tArg, Tobject)) tArgs)
    @ (match block with
      |None -> []
      |Some(Tproc(bArgs,bRet,None)) ->
	Less_than(bRet, Tobject)
	(* equal(bRet, Tobject) *)
	:: (List.map (fun bArg ->Less_than(bArg, Tobject)) bArgs)
      |Some typ -> invalid_arg ("unexpected block type: "^
    (Ikra_type.string_of typ)))
  | Has_method(Tclass c, any_method, Tproc(tArgs, tResult, block))
  (* | Has_method(Tclass c, any_method, tArgs, tResult, block) *)
    -> (* Printf.printf "deleting %s.%s" c any_method; *)
       Less_than(tResult,Tobject)
       (* equal(tResult,Tobject) *)
    :: (List.map (fun tArg -> Less_than(tArg, Tobject)) tArgs)
    @ (match block with
      |None -> []
      |Some(Tproc(bArgs,bRet,None)) ->
	equal(bRet, Tobject)
	:: (List.map (fun bArg ->Less_than(bArg, Tobject)) bArgs)
      |_ -> invalid_arg "unexpected")

  | Has_method(Tarray _, any_method, Tproc(tArgs, tResult, block))
    -> equal(tResult,Tobject)
    :: (List.map (fun tArg -> Less_than(tArg, Tobject)) tArgs)
    @ (match block with
      |None -> []
      |Some(Tproc(bArgs,bRet,None)) ->
	equal(bRet, Tobject)
	:: (List.map (fun bArg ->Less_than(bArg, Tobject)) bArgs)
      | _ -> invalid_arg "unexpected")

  | c -> [c]

let fourth_level_rules : craint_t
    -> (side_condition -> craint_t option) -> craint_t list =
  let open Type_constraints.Constr in
  let open Ikra_type in fun craint side_matcher ->
  let replace = replace craint in match craint with
  (* Int <: t1 => t1 = Int  ---  this rule will fix a type variable to
     the minimal type among the possibilities *)
  | Less_than(Tint,Tvar v) -> (* Printf.printf "third level rule makes t%d to int\n" v; *)
			      [equal(Tvar v,Tint)]
  (* t1 <: Float => t1 = Int --- this rule will fix a type variable to
  the minimal type among the possibilities *)
  | Less_than(Tvar v,Tfloat) -> (* Printf.printf "third level rule makes t%d to int\n" v; *)
    [equal(Tvar v,Tint)]

  (* Range <: t1 => t1 = Range,         t1 <: Range => t1 = Range *)
  | Less_than(Trange,t) | Less_than(t,Trange)
    -> replace "No sub/super for range"
    [equal(t,Trange)]

  (* Array(n,t0) <: Array(n,t1) => t0 = t1 *)
  (* | Less_than(Tarray(n,t0),Tarray(n',t1)) when n=n' -> [equal(t0,t1)] *)
  | Less_than(Tarray(n,t0),t1) | Less_than(t1,Tarray(n,t0))
    -> replace "No sub/super for array"
    [equal(Tarray(n,t0),t1)]

  | c -> [c]

let domain_specific_resolution : 'a -> (craints_t * subst_t) -> (craints_t * subst_t) =
  fun rules (craints,subst) ->
    let open Type_constraints.Constraints in
    let rec loop unprocessed processed =
      if is_empty unprocessed
      then (processed, subst)
      else let c = choose unprocessed in
           let unprocessed' = remove c unprocessed in
           let translated = rules c
             (fun sc ->
               try
                 Some (choose
                         (filter sc unprocessed'))
               with Not_found ->
                 try
		   Some (choose
                           (filter sc processed))
                 with Not_found -> None) in
           try 
           loop unprocessed'
             (List.fold_right add
                translated processed)
           with Type_constraints.Contradiction(contra) as e ->
             (Printf.printf "contradiction c=%s, contra=%s\n"
                (Type_constraints.Constr.string_of c)
                (Type_constraints.Constr.string_of contra)
               ; raise e) in
    loop craints empty 


let resolve_two : (craints_t * subst_t) -> (craints_t * subst_t) =
  domain_specific_resolution second_level_rules

let resolve_three : (craints_t * subst_t) -> (craints_t * subst_t) =
  domain_specific_resolution third_level_rules

let resolve_four : (craints_t * subst_t) -> (craints_t * subst_t) =
  domain_specific_resolution fourth_level_rules

let find_some : (('elt -> bool) -> 'set -> 'elt) -> ('elt -> 'b option) -> 'set
                -> 'b option

  (* find_some f ex s = Some v
     when there exists an element in s such that (ex element) = Some v;
     where f p s = e such that e is an element of s, and p e = true.
     If there is no such an element, None is returned.
  *)

  = fun find extract set ->
    try 
      (extract (find (fun elt -> Util.Option.is_some (extract elt)) set))
    with Not_found -> None

let rec type_variable_in_type : Ikra_type.t -> Ikra_type.t option =

  (* to extract a type variable in the given type; if there is no type
     variable, None is returned. *)

  let open Ikra_type in
  function
    | Tvar v -> Some(Tvar v)

    | Tint | Tfloat | Tbool | Trange | Tobject | Tclass _
      -> None

    | Tarray(_,t) | Tneighbor(_,_,t)
      -> type_variable_in_type t
    | Ttuple ts -> find_some List.find type_variable_in_type ts
    | Tproc(ts,_,_)
        when Util.Option.is_some (find_some List.find type_variable_in_type ts)
          -> find_some List.find type_variable_in_type ts
    | Tproc(_,t,_)
        when Util.Option.is_some (type_variable_in_type t)
          -> type_variable_in_type t
    | Tproc(_,_,Some t)
      -> type_variable_in_type t
    | Tproc(_,_,None)
      -> None

let type_variable_in_constraint : craint_t -> Ikra_type.t option =

  (* to extract a type variable in a given constraint; if there is no
     type variable, None is returned. *)

  let open Type_constraints.Constr in
  function
    |Less_than(t,_)  | Equal(t,_) |Has_method(t,_,_) | Defines_method(t,_,_)
        when (Util.Option.is_some (type_variable_in_type t)) ->
      type_variable_in_type t
        
    | Less_than(_,t) | Equal(_,t) | Has_method(_,_,t) |Defines_method(_,_,t)->
      type_variable_in_type t

    (* | _ -> None *)

let set_find : ('a -> bool) -> craints_t -> 'a =

  (* to pick a constraint in the given constrains such that satisfies
     pred *)

  fun pred set ->
    List.find pred (Type_constraints.Constraints.elements set)

let pick_type_variable : craints_t -> Ikra_type.t option =

  (* to pick one type variable in a given type constraint set.  None
     is returned if there is no type variable in the constraints.  *)

  fun craints ->
    find_some set_find type_variable_in_constraint craints

let eliminate_unknown_method_call : craints_t -> craints_t * craints_t =

  (* to eliminate a constraint of a method call with unknown method
     name as long as the return type is Object. *)
  let open Type_constraints.Constr in
  let open Ikra_type in
  Type_constraints.Constraints.partition (function
    | Has_method(recv_t, mname, Tproc(args_t, Tobject, None)) 
      -> true
    | _ -> false)

let resolve_five : (craints_t * subst_t) -> (craints_t * subst_t) =

  (* to add a constraint that equates a type variable in the given
     constraints to the Object type.  (NOTE: it only adds for one type
     variable; the other type variables remain unconstrained.  This
     will prevent to introduce contradiction.) *)

  let open Type_constraints.Constr in
      let open Ikra_type in
          fun (craints,subst) ->
    (* Printf.printf "LEVEL5: in %s,\n" (set_print craints); *)
    match pick_type_variable craints with
      |Some tvar -> (* Printf.printf "        force %s=Object\n" *)
        (* (Ikra_type.string_of tvar); *)
        Type_constraints.Constraints.add
        (equal(tvar, Tobject)) craints, subst
      |None      -> (* craints, subst *)
        (let unknown_calls, remaining =
           eliminate_unknown_method_call craints in
         if !show_trace then
           Printf.printf "LEVEL5: eliminates unknown calls %s,\n"
           (set_print unknown_calls)
         else ();
         remaining, subst
        )

let rec fix : ('a -> 'a) -> 'a -> 'a =
  fun f v ->
    let next_v = f v in
    if next_v = v
    then v
    else fix f next_v

let resolve_one' = 			(* just for testing purpoeses *)
  fun constraints ->
    (* print_string "resolve_one'\n"; *)
    fix resolve_one (constraints,empty_subst)

let wrap tag resolver =
  fun state ->
    if !show_trace then print_string ("begin "^tag^"\n") else ();
    (!trace state);
    let state' = resolver state in
    if !show_trace then print_string ("end "^tag^"\n") else ();
    (!trace state');
    state'

let resolve_one_two : (craints_t * subst_t) -> (craints_t * subst_t) =
  fun state ->
    (* (!trace state);  *)
    (fix (fun s -> (wrap "LEVEL2" resolve_two) ((fix resolve_one) s))) state

let resolve_one_to_three : (craints_t * subst_t) -> (craints_t * subst_t) =
  fun state ->
    (* (!trace state); (print_string "resolve_three\n"); *)
    (fix (fun s -> (wrap "LEVEL3" resolve_three) (fix resolve_one_two s))
       state)

let resolve_one_to_four : (craints_t * subst_t) -> (craints_t * subst_t) =
  fun state ->
    (* (!trace state); (print_string "resolve_four\n"); *)
    ((wrap "LEVEL4" resolve_four) (fix resolve_one_to_three state))

let resolve_one_to_five : (craints_t * subst_t) -> (craints_t * subst_t) =
  fun state ->
    (* (!trace state); (print_string "resolve_four\n"); *)
    ((wrap "LEVEL5" resolve_five) (fix resolve_one_to_four state))

let resolve : craints_t -> craints_t * subst_t =
  fun constraints -> 
    fix resolve_one_to_five (constraints,empty_subst)

exception Unresolved of Ast.tast * craints_t * Ast.tast * craints_t

(* TODO: add an optional parameter to lift unresolved type variables
   to object *)

let typeinf : Ast.tast -> Ast.tast =
  (* generates a type-annotated AST of the given program by applying a
     type inference algorithm *)

  fun program ->
    (* let constraints = Typeinf.gather Typeinf.init_env program in *)
    if !show_trace then
      Printf.printf "source:\n%s\n"
        (Ast.string_of ~show_type:true program)
    else ();
    let constraints = Typeinf.generate_constraints program in
    let (unresolved, subst) = resolve constraints in
    if (Type_constraints.Constraints.is_empty
	  (Type_constraints.remove_method_defs unresolved))
    then (apply_to_ast subst program)
    else let cs,s = resolve_one_two (constraints ,empty_subst) in
         raise (Unresolved(apply_to_ast s program, cs,
                           apply_to_ast subst program, unresolved))
(* invalid_arg("there are unresolved constraints: " *)
(* 		     ^ (set_print unresolved) ^ "\nin:" *)
(* 		     ^ (Ast.string_of ~show_type:true  *)
(* 			  (apply_to_ast subst program))) *)
