(* C/C++/CUDA expressions *)

(* http://www.cs.man.ac.uk/~pjj/bnf/c_syntax.bnf *)

type cuda_type = 
  | CUDAint
  | CUDAfloat
  | CUDAbool
  | CUDArange                           (* a pair of integer *)
  | CUDApointer of cuda_type * int list (* int list = the size of array *)
  | CUDAother of string
  | CUDAfunc of cuda_type * cuda_type list
  | CUDAtuple of cuda_type list
	| CUDAshared of cuda_type 
	| CUDAstatic of cuda_type

type variable = string
type vardecl = variable * cuda_type

type unary_op =
  | Make_ref				(* & *)
  | Deref				(* * *)
  | Unary_plus				(* + *)
  | Unary_minus				(* - *)
  | Bitwise_not				(* ~ *)
  | Logical_not				(* ! *)
  | Pre_inc				(* ++ *)
  | Pre_dec				(* -- *)
  (* | Post_inc				(\* ++ *\) *)
  (* | Post_dec				(\* -- *\) *)

type binary_op =
  | Logical_or				(* || *)
  | Logical_and				(* && *)
  | Bitwise_or				(* | *)
  | Exclusive_or			(* ^ *)
  | Bitwise_and				(* & *)
  | Equal				(* == *)
  (* | Not_equal				(\* != *\) *)
  | Less_than				(* < *)
  | Greater_than			(* > *)
  | Less_or_equal			(* <= *)
  | Greater_or_equal			(* >= *)
  | Shift_left				(* << *)
  | Shift_right				(* >> *)
  | Add					(* + *)
  | Sub					(* - *)
  | Mul					(* * *)
  | Div					(* / *)
  | Mod					(* % *)
  | Assign				(* = *)

type ternary_op =
  | Cond				(* ?: *)

type exp =
  | Literal of Ast.literal
  | Var of variable
  | Unary_op of unary_op * exp
  | Binary_op of binary_op * exp * exp
  | Ternary_op of ternary_op * exp * exp * exp
  | Statement_exp of statement list * exp
  | Funcall of string * exp list
  | Kernelcall of string * exp * exp * exp list
  | Cast of cuda_type * exp
  | Sizeof of cuda_type
  | Aref of exp * exp
  | Fref of exp * string

(* pointer -> *)

  (* Method def *)
  (* Yield *)
  (* Array *)

and statement =
  | Block of statement list
  | While of exp * statement
  | If of exp * statement * statement
  | For of statement * exp * exp * statement
  | Exp of exp
  | Return of exp
  | Return_
  (* | Break *)
  (* | Continue *)
  | Vardecl of vardecl
  | Vardecli of vardecl * exp
      (* do-while *)
      (* switch-case *)

type fundef = {
  modifier : string;
  return_type : cuda_type;
  name : string;
  formal_params : vardecl list;
  body : statement list;
}

type const_decl = {
  constant_modifier : string;
  constant_type : cuda_type;
  constant_name : string;
  constant_value : exp;
}

type struct_decl = {
  struct_name : string;
  struct_field_types : cuda_type list;
  struct_field_names : string list;
}

type decl =
  | Function_decl of fundef
  | Constant_decl of const_decl
  | Struct_decl of struct_decl

let unary_ruby_to_c = [
  "-@", Unary_minus, "-";
  "+@", Unary_plus, "+";
  "not@", Logical_not, "!";
  "~", Bitwise_not, "~";
  "****",Pre_inc, "++";	
  "****",Pre_dec, "--";	
  "&",   Make_ref, "&";	
  (* "****",Post_inc, "++"; *)
  (* "****",Post_dec, "--"; *)
]
 
let binary_ruby_to_c = [
  "||", Logical_or,      "||";
  "&&", Logical_and,     "&&";
  "|",  Bitwise_or,	  "|";
  "^",  Exclusive_or,	  "^";
  "&",  Bitwise_and,	  "&";
  "==", Equal,		  "==";
  (* "!=", Not_equal,	  "!="; *)
  "<",  Less_than,	  "<";
  ">",  Greater_than,	  ">";
  "<=", Less_or_equal,	  "<=";
  ">=", Greater_or_equal, ">=";
  "<<", Shift_left,	  "<<";
  ">>", Shift_right,	  ">>";
  "+",  Add,		  "+";
  "-",  Sub,		  "-";
  "*",  Mul,		  "*";
  "/",  Div,		  "/";
  "%",  Mod,		  "%";
  "=",  Assign,		  "=";
]


let constant_table : (string * string) list =
  (* to define a list of Ruby constants that are defined in the CXX
     namespace. *)
  [
    "Math",             "";
    "Math::PI",         "M_PI";
    "true",             "TRUE";
    "false",            "FALSE";
  ]

let is_cxx_defined_constant : string -> bool =
  fun cname -> List.mem_assoc cname constant_table

let lookup_op : ('b*'a*string) list -> string ->  'a =

  fun table search_name ->
    let _,op,_ =
      List.find (fun (name,_,_) -> name=search_name) table in
    op


let lookup_unary_op : string -> unary_op =

  lookup_op unary_ruby_to_c


let method_name_is_unary : string -> bool =
  fun search_name ->
    try let _ = lookup_unary_op search_name in
	true
    with Not_found -> false

let lookup_binary_op : string -> binary_op =

  lookup_op binary_ruby_to_c

let method_name_is_binary : string -> bool =
  fun search_name ->
    try let _ = lookup_binary_op search_name in
	true
    with Not_found -> false

(* let c_op_exp : string -> exp list -> exp = *)
(*   fun ruby_name operands -> match (ruby_name, operands) with *)
(*     | *)

let rec type_signature_of_ikra_type : cuda_type -> string =
  (* given an Ikra type, encode the type into a string like JVM does.
     A tuple type is encoded into a number of elements followed by its
     element types in encoded representations.

     e.g.,
     int -> "I"
     float -> "F"
     int * float -> 2IF
  *)
  function
  | CUDAint -> "I"
  | CUDAfloat -> "F"
  | CUDAbool -> "B"
  | CUDAtuple ts -> (string_of_int (List.length ts)) ^
      String.concat "" (List.map type_signature_of_ikra_type ts)
  | _ -> invalid_arg "unsupported"
      
    
let rec cuda_type_of_ikra_type : Ikra_type.t -> cuda_type =
  let open Ikra_type in
  function
    | Tint -> CUDAint
    | Tfloat -> CUDAfloat
    | Tbool  -> CUDAbool
    | Tarray(_,element_type) ->
      CUDApointer(cuda_type_of_ikra_type element_type,[])
    | Trange -> CUDArange
    (* TODO: Tneighbor *)
    | Ttuple ts -> CUDAtuple(List.map cuda_type_of_ikra_type ts)
    | t -> invalid_arg ("no CUDA type for: "^(Ikra_type.string_of t))

let constructor_name_of_cuda_tuple : cuda_type -> string = function
  | (CUDAtuple _) as t -> "ikra_tuple_" ^(type_signature_of_ikra_type t)
  | _ -> invalid_arg "unsupported"

exception Has_statement

let statement_of_type_decl : Ast.type_decl -> statement =
  fun (var,typ) ->
    Vardecl(var, cuda_type_of_ikra_type typ)

let rec exp_of_ruby : Ast.tast -> exp =
  (* exp_of_ruby ruby_expression constructs a CUDA expression for the
     given ruby_expression.  When the ruby_expression cannot be
     expressed as a C++/CUDA expression (e.g., a while expression), it
     raises Invalid_argument. *)

  function
    (* binary operator specific to a type *)
    | Ast.Call((_,typ) as exp1, "%", [exp2], None),_ as e->
      if Ikra_type.Tfloat = typ 
      then Funcall("fmodf", [exp_of_ruby exp1; exp_of_ruby exp2])
      else if Ikra_type.Tint = typ 
      then Binary_op (lookup_binary_op "%", exp_of_ruby exp1, exp_of_ruby exp2)
      else invalid_arg ("neither float nor int:"
                        ^(Ast.string_of ~show_type:true e))
    (* binary operator *)
    | Ast.Call(exp1, method_name, [exp2], None),_
      when method_name_is_binary method_name ->
      Binary_op (lookup_binary_op method_name,
		 exp_of_ruby exp1,
		 exp_of_ruby exp2)

    (* TODO: this is too ad-hoc *)
    | Ast.Call(exp1, "**", [exp2], None),_->
      Funcall("fexp",[exp_of_ruby exp1; exp_of_ruby exp2])

    (* unary operator *)
    | Ast.Call(exp1, method_name, [], None),_
      when method_name_is_unary method_name ->
      Unary_op (lookup_unary_op method_name,
		exp_of_ruby exp1)

    (* unary operator *)
    | Ast.Call(exp1, method_name, [], None),_
      when List.mem method_name ["abs";"floor"] ->
      Funcall(List.assoc method_name ["abs","fabs"; "floor","floorf";],
              [exp_of_ruby exp1])

    (* range operator, e1..e2, which is represented as a call of
       __PSEUDO_DOT2 on self.  This will be converted to a constructor
       call on ruby_range structure. *)
    | Ast.Call((* (_,Ikra_type.Tobject) as *) reciever,
               method_name, exps, None),_
      when Ast.is_self reciever && method_name = "__PSEUDO_dot2" ->
      Funcall("_Ikra_range", List.map exp_of_ruby exps)

    (* TODO: generalize this *)
    (* logical not *)
    | Ast.Call((_,Ikra_type.Tobject) as reciever, "not@", [exp1], None),_
      when Ast.is_self reciever ->
      Unary_op (lookup_unary_op "not@",
		exp_of_ruby exp1)

    (* logical or and and *)
    | Ast.Call((_,Ikra_type.Tobject) as reciever, op, [exp1;exp2], None),_
      when Ast.is_self reciever &&
        List.mem op ["||";"&&";] ->
      Binary_op (lookup_binary_op op,
		 exp_of_ruby exp1,
		 exp_of_ruby exp2)

    (* method call on self; i.e., global function *)
    | Ast.Call((_,Ikra_type.Tobject) as reciever, method_name, exps, None),_
      when Ast.is_self reciever ->
      Funcall(method_name, List.map exp_of_ruby exps)


    (* TODO: in case of higher dimensional tuple *)
    | Ast.Call(exp1, "[]", [exp2], None),_ ->
      Aref(exp_of_ruby exp1, exp_of_ruby exp2)

    (* built-in functions *)
    (* TODO: list more cases *)
    | Ast.Call((Ast.Constant("Math"),_), mname, args, None),_
      when List.mem mname ["acos";"acosh";"asin";"asinh"; "atan"; "atanh";
                           "cbrt"; "cos"; "cosh"; "erf"; "erfc"; "exp";
                           "gamma"; "lgamma"; "log"; "log10"; "log2";
                           "rsqrt"; "sin"; "sinh"; "sqrt"; "sqrt"; "tan";
                           "tanh";] ->
      Funcall(mname, List.map exp_of_ruby args)

    (* cast *)
    | Ast.Call(exp1, "to_i", [], None),_  ->
      Cast(CUDAint, exp_of_ruby exp1)

    (* literal  *)
    | Ast.Literal(literal),_ -> Literal(literal)
    (* constant  *)
    | Ast.Constant(name),_  ->
      Var(try List.assoc name constant_table
        with Not_found -> name)     (* TODO: make this name safe by adding
                                       some prefix like "IKRA_CONST$" *)

    (* assignment to local variable *)
    | Ast.Local_assign (name,exp),_  ->
      Binary_op(Assign, Var(name), exp_of_ruby exp)
    (* local variable reference *)
    | Ast.Local_var (name),_  -> Var name
    (* if *)
    | Ast.If(cond,then_exp,Some else_exp),_ ->
      Ternary_op(Cond, exp_of_ruby cond,
		 exp_of_ruby then_exp, exp_of_ruby else_exp)
    (* NOTE: IF-THEN-without-ELSE cannot be an expression *)
    | Ast.If(_,_,None),_ -> raise Has_statement
    | Ast.While _,_ -> raise Has_statement
    | Ast.For _,_ -> raise Has_statement

    (* block statements *)
    | Ast.Block(exps),_ ->
      let butlast, last = Util.split_last exps in
      Statement_exp(List.map statement_of_ruby_ast butlast,
                    exp_of_ruby last)

    (* | Ast.Type_decl(_decl, [unique_body]),_ -> exp_of_ruby unique_body *)
    | Ast.Type_decl(decls, body_exps),_ ->
      let butlast, last = Util.split_last body_exps in
      Statement_exp(
	(List.map statement_of_type_decl decls)
	@ (List.map statement_of_ruby_ast butlast),
        (exp_of_ruby last))
    (* [e0,...,en] -> ikra_tuple_T##(e0',...,en') where ## is type signature *)
    | Ast.Array(exps),t ->
      let c = constructor_name_of_cuda_tuple (cuda_type_of_ikra_type t) in
      Funcall(c, List.map exp_of_ruby exps)
    | e -> invalid_arg ("unexpected type of expression:"^ (Ast.mlexp_of Ikra_type.string_of e))

and statement_of_ruby_ast : Ast.tast -> statement =
  (* returns_of_ruby_ast e generates a list of CUDA statements that
     contains return statements at its tail position. *)

  let block : statement list -> statement =
    (* block statements constructs a Block node with flattening nested
       blocks in the given statements *)
    
    let rec flatten ss =
      List.fold_left (@) []
	(List.map
	   (function | Block(sss) -> flatten sss | s -> [s])
	   ss) in
    
    fun ss -> Block(flatten ss) in


  function
    (* method call *)
    | Ast.Call _,_ as e -> Exp (exp_of_ruby e)
    (* literal  *)
    | Ast.Literal(literal),_ as e -> Exp (exp_of_ruby e)
    (* constant  *)
    | Ast.Constant(name),_  as e ->  Exp (exp_of_ruby e)
    (* assignment to local variable *)
    | Ast.Local_assign (name,ast),_  as e -> Exp (exp_of_ruby e)
    (* local variable reference *)
    | Ast.Local_var (name),_  as e -> Exp (exp_of_ruby e)
    (* block structure*)
    | Ast.Block (body),_  ->
      block(List.map statement_of_ruby_ast body)
    (* while loop *)
    (* | Ast.While (cond, body),_  ->  *)
    (* conditional branch *)
    | Ast.If (cond, then_exp, Some else_exp),_ as e ->
      (try Exp (exp_of_ruby e)
       with Has_statement -> 
	 If(exp_of_ruby cond,
	    statement_of_ruby_ast then_exp,
	    statement_of_ruby_ast else_exp))
    | Ast.If (cond, then_exp, None),_ as e ->
      (try Exp (exp_of_ruby e)
       with Has_statement -> 
	 If(exp_of_ruby cond,
	    statement_of_ruby_ast then_exp,
	    block []))

    (* For(v,e1,e2) is converted to

       range tmp_range = [[e1]];
       for (T v = tmp_range.min; v < tmp_range.max; v++) [[e2]]

       where range is defined like this

       typedef struct _ikra_range {
       int min, max;
       range(int _min, int _max): min(_min), max(_max) {}
       } _ikra_range_t;

       
    *)

    (* for loop *)
    | Ast.For (vname, range, body),_  ->
      (* TODO: don't use Cuda_constructors here, or split this file
         into two to avoid circular dependency. *)
      let rvar = "_ikra_tmp_range" in
      Block
        [Vardecli((rvar, CUDArange),exp_of_ruby range
                  (* Binary_op(Assign, Var vname, exp_of_ruby range) *));
         For(Vardecli((vname, CUDAint), Fref(Var rvar, "min")),
             Binary_op(Less_or_equal, Var vname, Fref(Var rvar, "max")),
             Unary_op(Pre_inc, Var vname),
             statement_of_ruby_ast body)]

    (* return *)
    | Ast.Return(Some exp),_ -> Return(exp_of_ruby exp)
    | Ast.Return(None),_     -> Return_

    (* type_decl --- only after type inference *)
    | Ast.Type_decl (tdecls, body),_ -> 
      block(
	(List.map (fun (v,typ) ->
	  Vardecl(v, cuda_type_of_ikra_type typ))
	   tdecls) @
	  (List.map statement_of_ruby_ast body))

    (* while *)
    | Ast.While (cond,body),_ ->
      While(exp_of_ruby cond, statement_of_ruby_ast body)

    (* unsupported constructs *)

    (* procedure as a parameter *)
    | Ast.Proc _,_
    (* method definition *)
    | Ast.Method_def _,_ 
    (* yield *)
    | Ast.Yield _,_ 
    (* array literal *)
    | Ast.Array _,_  
    (* constant assignment *)
    | Ast.Constant_assign _,_
      as e  ->
      invalid_arg ("unsupported now:" ^ (Ast.mlexp_of Ikra_type.string_of e))

let decl_of_ruby_ast : Ast.tast -> decl =
  let f mname params bexps typ def = 
    (try
       let formal_params = (match params with
         | Ast.Array_pattern ps ->
           List.map (function
             | Ast.Var_pattern (v,t) -> (v,cuda_type_of_ikra_type t)
             | _ -> invalid_arg "unexpected parameter (nested?)")
             ps
         | _ -> invalid_arg "unexpected parameter pattern") in
       let bexp' = match bexps with
	 | [Ast.Type_decl(decls,body),t] ->
	   Ast.Type_decl(List.filter
			   (fun (v,_) -> not(List.mem_assoc v formal_params)) 
			   decls,
		         body),t
         | other -> Ast.Block(other),typ
           (* invalid_arg "block body has not type_decl" *) in
       Function_decl 
         {
	   modifier = "";
           return_type = cuda_type_of_ikra_type typ;
           name = mname;
           formal_params = formal_params;
           body = [Return (exp_of_ruby bexp')];
         }
     with ex -> (Printf.printf "error while converting %s\n"
                   (Ast.string_of ~show_type:true def);
                 raise ex)) in
  (* TODO: clean this function.  first, the syntax for method
     definitions should be more uniform.  Second, the pattern machings
     in the above and below are somewhat redundant. *)

  (* FIXME: do not use the type of the last statement of the body as
     the return type of the method, as it can be different when there is
     a return statement in the middle of the body.  For example,

     def m()
       return 0.5
       123
     end

     should have return type of float, but the last statement has a
     type of integer.
  *)
  function
    | Ast.Method_def(mname, params,
                     ((Ast.Type_decl _,typ) as bexp)), _ as def ->
      f mname params [bexp] typ def
    | Ast.Method_def(mname, params,
                     (Ast.Block(bexps),typ)), _ as def->
      f mname params bexps typ def
    | Ast.Constant_assign(constant_name, exp1), typ ->
      Constant_decl {
        constant_modifier = "";
        constant_type = cuda_type_of_ikra_type typ;
        constant_name;
        constant_value = exp_of_ruby exp1;
      }
    | exp -> invalid_arg ("unsupported type of expression:"
                          ^ ((Ast.mlexp_of Ikra_type.string_of) exp))
      
let struct_name_of_cuda_tuple : cuda_type -> string = function
  | (CUDAtuple _) as t -> "ikra_tuple_" ^(type_signature_of_ikra_type t)
  | _ -> invalid_arg "unsupported"

(* let rec string_of_cuda_type : cuda_type -> string = *)
(*   function *)
(*     | CUDAint -> "int" *)
(*     | CUDAfloat -> "float" *)
(*     | CUDAbool -> "boolean" *)
(*     | CUDArange -> "ikra_range_t" *)
(*     | CUDApointer t ->  (string_of_cuda_type t) ^"*" *)
(*     | CUDAother t -> t *)
(*     | CUDAfunc(ret_t, arg_t) -> *)
(*       Printf.sprintf "%s(\*\)(%s)" *)
(* 	(string_of_cuda_type ret_t) *)
(* 	(String.concat "," (List.map string_of_cuda_type arg_t)) *)
(*     | (CUDAtuple _) as t -> "struct " ^ (struct_name_of_cuda_tuple t) *)

let string_of_array_length : int list -> string =
	fun len_list ->
	List.fold_left (fun v1 v2 -> v1 ^ v2) "" (List.map (fun v -> "[" ^ (string_of_int v) ^ "]") len_list)
								 
let rec string_of_cuda_type : ?vname:string -> cuda_type -> string =
	fun ?(vname="") ->
	function
    | CUDAint -> "int"
    | CUDAfloat -> "float"
    | CUDAbool -> "boolean"
    | CUDArange -> "ikra_range_t"
    | CUDApointer (t,len_lst) ->
			 ( match len_lst with
					 | [] -> (string_of_cuda_type t) ^"*"
					 | _ -> (string_of_cuda_type t) ^ " " ^ vname ^ (string_of_array_length len_lst)
			 )
    | CUDAother t -> t
    | CUDAfunc(ret_t, arg_t) ->
      Printf.sprintf "%s(*)(%s)"
	(string_of_cuda_type ret_t)
	(String.concat "," (List.map string_of_cuda_type arg_t))
    | (CUDAtuple _) as t -> "struct " ^ (struct_name_of_cuda_tuple t)
		| CUDAshared t -> "__shared__ " ^ (string_of_cuda_type ~vname:vname t)
		| CUDAstatic t -> "static " ^ (string_of_cuda_type ~vname:vname t)

let rec is_array_decl : cuda_type -> bool =
	function
	| CUDApointer (t,len_list) ->
		 ( match len_list with
				 | [] -> false
				 | _ -> true
		 )
	| CUDAshared t -> is_array_decl t
	| CUDAstatic t -> is_array_decl t
	| _ -> false



let string_of_op : ('b*'a*string) list -> 'a -> string =

  fun table search_op ->
    let (_,_,name) = List.find (fun (_,op,_) -> op = search_op ) table in
    name

let string_of_unary = string_of_op unary_ruby_to_c
let string_of_binary = string_of_op binary_ruby_to_c

let rec string_of_exp : exp -> string =
  
  let par exp = Util.in_paren (string_of_exp exp) in

  function
    | Literal(Ast.Integer_literal n) -> string_of_int n
    | Literal(Ast.Float_literal n) -> string_of_float n
    | Literal(Ast.String_literal s) -> "\"" ^ s ^ "\""
    | Literal(l)  -> invalid_arg ("unsupported literal:" ^
				     (Ast.string_of_literal l))
    | Var v -> v
    | Unary_op(op,exp) -> (string_of_unary op) ^ (par exp)
    | Binary_op(op,exp1,exp2) ->
      (par exp1) ^ (string_of_binary op) ^ (par exp2)
    | Ternary_op(Cond,exp1,exp2,exp3) ->
      (par exp1) ^ "?" ^ (par exp2) ^ ":" ^ (par exp3)
    | Statement_exp (statements,exp) ->
      "({"
      ^ (String.concat "" (List.map string_of_statement statements))
      ^ (string_of_exp exp) 
      ^";})"
    | Funcall(fname,exps) ->
      fname ^ "(" ^ (String.concat "," (List.map string_of_exp exps)) ^ ")"

    | Kernelcall(fname,num_block,num_threads,exps) ->
      Printf.sprintf "%s<<<%s,%s>>>(%s)"
        fname (string_of_exp num_block) (string_of_exp num_threads)
        (String.concat "," (List.map string_of_exp exps))
    | Aref(exp1,exp2) ->
      (string_of_exp exp1) ^ "[" ^ (string_of_exp exp2) ^ "]"
    | Cast(typ,exp) ->
      Printf.sprintf "(%s)(%s)" (string_of_cuda_type typ)
	(string_of_exp exp)
    | Sizeof(typ) ->
      Printf.sprintf "sizeof(%s)" (string_of_cuda_type typ)
    | Fref(exp,fname) ->
      Printf.sprintf "(%s).%s" (string_of_exp exp) fname

and string_of_statement : statement -> string =
  function
  | Block statements ->
    "{"^(String.concat "" (List.map string_of_statement statements))^"}"
  | While(exp,statement) ->
    "while("^(string_of_exp exp)^")"^(string_of_statement statement)
  | If(exp,s1,s2) ->
    "if("^(string_of_exp exp)^")"^(string_of_statement s1)^" else "
      ^(string_of_statement s2)
  | For(init,cond,cont,body) ->
    Printf.sprintf "for(%s%s;%s)%s"
      (string_of_statement init)
      (string_of_exp cond)
      (string_of_exp cont)
      (string_of_statement body)
  | Exp exp -> (string_of_exp exp) ^ ";"
  | Return exp -> "return " ^ (string_of_exp exp) ^ ";"
  | Return_  -> "return;"
  (* | Break *)
  (* | Continue *)
  | Vardecl(var,typ) ->
		 if (is_array_decl typ)
		 then
			 (string_of_cuda_type ~vname:var typ) ^ ";"
		 else
			 (string_of_cuda_type typ) ^ " " ^ var ^ ";"
  | Vardecli((var,typ),exp) ->
    (string_of_cuda_type typ) ^ " " ^ var ^ "=" ^ (string_of_exp exp) ^";"
      
let string_of_statements =
  fun ss -> String.concat "" (List.map string_of_statement ss)
  
let prototype_of_decl : decl -> string = function
  | Function_decl {modifier; return_type; name; formal_params} ->
    modifier
    ^ (string_of_cuda_type return_type) ^ " "
    ^ name
    ^ "(" ^ (String.concat ","
	       (List.map (fun (v,t) -> (string_of_cuda_type t) ^ " " ^ v)
		  formal_params))
    ^")"
  | Constant_decl _ -> ""

let prepend_modifier : string -> decl -> decl =
  fun additional_modifier -> function
  | Function_decl{ modifier; return_type; name; formal_params; body; }
    -> Function_decl{ modifier = additional_modifier ^ " " ^ modifier;
         return_type; name; formal_params; body; }
  | Constant_decl{ constant_modifier;
                   constant_type; constant_name; constant_value; }
    -> Constant_decl{ constant_modifier
                      = additional_modifier ^ " " ^ constant_modifier;
                      constant_type; constant_name; constant_value; }


let string_of_decl : decl -> string = function
  | Function_decl {body=body} as def ->
    (prototype_of_decl def)
    ^ "{\n" ^ (String.concat "\n"
		 (List.map string_of_statement body)) ^ "\n}"
  | Constant_decl {constant_type; constant_name; constant_value} ->
    "const " ^
    (string_of_cuda_type constant_type) ^ " "
    ^ constant_name ^ " = "
    ^ (string_of_exp constant_value) ^ ";"
  | Struct_decl {struct_name; struct_field_types; struct_field_names} ->
    "struct " ^ struct_name ^ " {\n" ^
      (String.concat "\n"
         (List.map2 (fun ftype fname ->
           "\t" ^ (string_of_cuda_type ftype) ^ " " ^ fname ^ ";")
            struct_field_types struct_field_names)) ^ "\n};\n"
         
             
             
