(* smart constructors for C/C++/CUDA expressions *)

open Cuda

module Helper = struct
  (* those are not used by customers *)

  let make_uniop : unary_op -> (exp -> exp) =
    fun op e1 -> Unary_op(op,e1)

  let make_binop : binary_op -> (exp -> exp -> exp) =
    fun op e1 e2 -> Binary_op(op,e1,e2)

  let counter = ref 0
  let generate_var_name _ =
    let c = !counter in
    (counter := c+1;
     Printf.sprintf "v%05d" c)

  let vargen : ?name:string -> cuda_type -> (vardecl * exp) =
    fun ?(name=generate_var_name()) typ -> ((name,typ), Var name)

end

(* cuda_type *)
let _int   = CUDAint
let _float = CUDAfloat
(* let _p typ = CUDApointer(typ) *)
let void   = CUDAother "void"

(* exp *)
let _num n = Literal (Ast.Integer_literal n)
let _flt n = Literal (Ast.Float_literal n)
let _str n = Literal (Ast.String_literal n)
(* Var --- use vargen *)
(* Unary_op --- at the bottom *)
(* Binary_op --- at the bottom *)
let _ifexp e1 e2 e3 = Ternary_op(Cond,e1,e2,e3)
let _statement statements exp = Statement_exp(statements,exp)
let ( **@ ) fname args = Funcall(fname,args)
let _aref e1 e2 = Aref(e1,e2)
let _cast t e = Cast(t,e)
let sizeof t = Sizeof(t)
let _fref e f = Fref(e,f)

(* statement *)
let _block statements = Block statements
let _while exp statement = While(exp,statement)
let _if cond_exp then_st else_st = If(cond_exp,then_st,else_st)
let _for : cuda_type -> string ->
  (exp -> (exp * exp * exp * statement)) -> statement =
  fun typ vname forgen ->
    let (vdecl, var) = Helper.vargen ~name:vname typ in
    let (init, cond, inc, body) = forgen var in
    For(Vardecli(vdecl,init), cond, inc, body)
let _exp e = Exp e			
let return exp = Return exp
let return_ = Return_
let _let : cuda_type -> string -> exp -> (exp -> statement list)
  -> statement list =
  fun typ vname exp body_gen ->
    let (vdecl, var) = Helper.vargen ~name:vname typ in
    (Vardecli(vdecl,exp)) :: (body_gen var)
let _let0 : cuda_type -> string -> (exp -> statement list)
  -> statement list =
  fun typ vname body_gen ->
    let (vdecl, var) = Helper.vargen ~name:vname typ in
    (Vardecl(vdecl)) :: (body_gen var)

let ikra_check_kernel_call () = "ikra_check_kernel_call" **@ []
let _kernel_call fname (num_block,num_threads) args =
  [ _exp (Kernelcall(fname,num_block,num_threads,args));
    _exp (ikra_check_kernel_call()); ]
(* fundef *)

let defun : ?modifier:string -> cuda_type -> string -> (cuda_type * string) list
  -> (exp list -> statement list) -> fundef =
  fun ?(modifier="") return_type name types_vars body_gen ->
    {
      modifier;
      return_type;
      name;
      formal_params = List.map (fun (typ,name) -> (name,typ)) types_vars;
      body = body_gen (List.map (fun (_,name) -> Var name) types_vars);
  }

let inline = defun ~modifier:"inline "
let __global__ = defun ~modifier:"__global__\n"
let __device__ = defun ~modifier:"__device__\n"
let __device__inline = defun ~modifier:"__device__\ninline "
let __device__host__inline = defun ~modifier:"__device__ __host__\ninline "
let extern_C = defun ~modifier:"extern \"C\"\n"

let _struct : string -> (cuda_type * string) list -> struct_decl =
  fun struct_name field_specs ->
    {
      struct_name;
      struct_field_types = List.map (fun (typ,name) -> typ ) field_specs;
      struct_field_names = List.map (fun (typ,name) -> name) field_specs;
    }

(* ********************************************************************** *)
(* predefined symbols *)

let _VALUE = CUDAother "VALUE"
let _VALUE_func_t = CUDAfunc(_VALUE, [CUDAother "..."])
let get_parray (wrapped_array, typ) =
  "get_parray" **@ [ wrapped_array; Var (Cuda.string_of_cuda_type typ)]
let rb_int_new exp = "rb_int_new"**@[ exp ]
let rb_float_new exp = "rb_float_new"**@[ exp ]
let rb_define_module_function(exp1, exp2, exp3, exp4) =
  "rb_define_module_function" **@ [exp1; exp2; exp3; exp4]
let rb_define_module(exp) = "rb_define_module" **@ [exp]
let _NUM2INT arg = "NUM2INT" **@ [arg]
let _IKRA_NUM2FLOAT arg = "IKRA_NUM2FLOAT" **@ [arg]
let _Qnil = Var "Qnil"

let _range_min exp = _fref exp "min"
let _range_max exp = _fref exp "max"

(* let amp = Helper.make_uniop Make_ref *)

(* CUDA specific  *)
let _NUM_THREADS = Var "NUM_THREADS"
let blockIdx_x = Var "blockIdx.x"
let blockDim_x = Var "blockDim.x"
let threadIdx_x = Var "threadIdx.x"
let gridDim_x = Var "gridDim.x"
let blockIdx_y = Var "blockIdx.y"
let blockDim_y = Var "blockDim.y"
let threadIdx_y = Var "threadIdx.y"
let blockIdx_z = Var "blockIdx.z"
let blockDim_z = Var "blockDim.z"
let threadIdx_z = Var "threadIdx.z"

type cudaMemcpyKind =
  | CudaMemcpyDeviceToHost | CudaMemcpyHostToDevice | CudaMemcpyDeviceToDevice

let cudaMemcpyKindVar = function
  | CudaMemcpyDeviceToHost -> Var "cudaMemcpyDeviceToHost"
  | CudaMemcpyHostToDevice -> Var "cudaMemcpyHostToDevice"
  | CudaMemcpyDeviceToDevice -> Var "cudaMemcpyDeviceToDevice"

(* TODO: call wrapper functions that check the results. *)
let cudaMemcpyDeviceToHost = Var "cudaMemcpyDeviceToHost"
let cudaMemcpy dst src nbytes direction =
  "ikra_memcpy" **@ [dst; src; nbytes; cudaMemcpyKindVar direction]
let cudaMalloc ptr_ptr size = "ikra_alloc"**@[ptr_ptr; size]
let cudaFree ptr = "ikra_free"**@[ptr]

(* ********************************************************************** *)
(* overload operators 

   These are defined by following the rules so as not to override
   existing OCaml operators:

   1. type level prefix operators begin with "?"
   2. expression level prefix operators begin with "!" 
   3. expression level infix operators begin with "$"
      (which will be left-associative).
*)

let ( ?*  ) = fun typ -> CUDApointer (typ,[])
          
let ( !&  ) = Helper.make_uniop Make_ref
let ( !*  ) = Helper.make_uniop Deref
let ( !+  ) = Helper.make_uniop Unary_plus
let ( !-  ) = Helper.make_uniop Unary_minus
let ( !~  ) = Helper.make_uniop Bitwise_not
let ( !!  ) = Helper.make_uniop Logical_not
let ( !++ ) = Helper.make_uniop Pre_inc
let ( !-- ) = Helper.make_uniop Pre_dec	

let ( ||@ ) = Helper.make_binop Logical_or 
let ( &&@ ) = Helper.make_binop Logical_and 
let ( |@  ) = Helper.make_binop Bitwise_or
let ( ^@  ) = Helper.make_binop Exclusive_or 
let ( &@  ) = Helper.make_binop Bitwise_and 
let ( ==@ ) = Helper.make_binop Equal 
let ( <@  ) = Helper.make_binop Less_than 
let ( >@  ) = Helper.make_binop Greater_than 
let ( <=@ ) = Helper.make_binop Less_or_equal 
let ( >=@ ) = Helper.make_binop Greater_or_equal 
let ( <<@ ) = Helper.make_binop Shift_left 
let ( >>@ ) = Helper.make_binop Shift_right 
let ( +@  ) = Helper.make_binop Add 
let ( -@  ) = Helper.make_binop Sub 
let ( *@  ) = Helper.make_binop Mul 
let ( /@  ) = Helper.make_binop Div 
let ( %@  ) = Helper.make_binop Mod 
let ( =@  ) = Helper.make_binop Assign 
