(* frontend.ml *)

let parse_error_handler source_file = function
  | Util.External_command_error (command, stdout, stderr) as e ->
  Printf.printf "Parse error while reading %s.\n\
 Parser command: %s\n\
 Parser output:\n%s\n\
 Parser error:\n%s\n" source_file command stdout stderr;
  raise e
  | _ -> invalid_arg "never happens"

let type_error_handler = function
  | Cresolve.Unresolved(ast2, raints2, ast3, raints3) as e ->
  Printf.printf "Type inferencing failed.\n\
 Lv2: source=%s\n\
 Lv2: constraints=%s\n
 Lv3: source=%s\n\
 Lv3: constraints=%s\n"
    (Ast.string_of ~show_type:true ast2)
    (Cresolve.set_print raints2)
    (Ast.string_of ~show_type:true ast3)
    (Cresolve.set_print raints3);
  raise e
  | _ -> invalid_arg "never happens"

let generate_error_handler typed_ast ex =
  print_string "Error while generate_and_compile for\n";
  print_string (Ast.string_of ~show_type:true typed_ast);
  raise ex 

let execution_error_handler gen_env ex =
  let ruby_file = (Convert_env.get_ruby_file_name gen_env) in
  let cuda_file = (Convert_env.get_cuda_file_name gen_env) in
  Printf.printf "Error while running a ruby program (%s):\n%s\nand a CUDA program (%s):\n%s\n"
    ruby_file (Util.read_lines ruby_file)
    cuda_file (Util.read_lines cuda_file);
  raise ex

(* backend chooser ***************************************************)
type 'a backend = {
  target_file_names
  : ?cleanup:bool -> string option -> string -> (Convert_env.t -> 'a) -> 'a;
  generate_and_compile
    : ?generate_only:bool -> Convert_env.t -> Ast.tast -> unit;
  run_converted : ?generate_only:bool -> ?redirect:bool -> Convert_env.t
               -> string list -> string;
}
    
module ConvertCPP = Convert.M(Convert_kernel)
let cpp_backend = {
  target_file_names = ConvertCPP.target_file_names;
  generate_and_compile = ConvertCPP.generate_and_compile;
  run_converted = ConvertCPP.run_converted;
}
module ConvertCUDA = Convert.M(Convert_kernel_cuda)
let cuda_backend = {
  target_file_names = ConvertCUDA.target_file_names;
  generate_and_compile = ConvertCUDA.generate_and_compile;
  run_converted = ConvertCUDA.run_converted;
}
let backends = [
  "cpp",  cpp_backend;
  "cuda", cuda_backend;                 (* TODO: which this is chosen, the suffix of generated cxx file should become ".cu", rather than ".cc" *)
(* TODO: when this is chosen, the makefile should use nvcc rather than
   g++. *)
]

let find_backend : string -> 'a backend =
  fun name -> (* Printf.printf "\nbackend=%s\n" name; *)
    List.assoc name backends
(* ***************************************************)


let compile_file : ?cleanup:bool -> ?generate_only:bool -> ?backend:string
  -> ?target_dir:(string option) -> ?redirect:bool -> string -> string list
  -> string =
  (* generate compiled files for the given source_file, and returns the
     file name of the generated ruby script.  *)

  fun ?(cleanup=true) ?(generate_only=false) ?(backend="cpp")
    ?(target_dir=None) ?(redirect=false) source_file app_args ->
    let backend = find_backend backend in
    let source_ast = try Parser.parse_file source_file
      with (Util.External_command_error _) as e ->
        parse_error_handler source_file e in
    let ast_with_let = Let_intro.let_intro source_ast in
    let ast_with_type_vars = Typeinf.add_type_vars ast_with_let in
    let typed_ast = try Cresolve.typeinf ast_with_type_vars
      with (Cresolve.Unresolved _) as e -> type_error_handler e in
    (* Printf.printf "typed source:\n%s\n\n" *)
    (*   (Ast.string_of ~show_type:true typed_ast); *)
    backend.target_file_names ~cleanup target_dir source_file (fun gen_env ->
      let _ = try backend.generate_and_compile ~generate_only gen_env typed_ast 
	with ex -> generate_error_handler typed_ast ex in
      try 
	let r = backend.run_converted ~generate_only
          ~redirect gen_env app_args in
        if generate_only then
          let r = Convert_env.get_ruby_file_name gen_env in
          print_string "\n"; print_string r; print_string "\n"; r
        else r
      with ex -> execution_error_handler gen_env ex;
    )

let compile_code : string -> string =
  (* compiles an Ikra program given as the program_string, and returns the
     file name of the generated ruby script. *)

  fun program_string ->
    let source_file_name, outchan =
      Filename.open_temp_file "ikra_source" ".rb" in
    Util.with_finally (fun () ->
      let _ = output_string outchan program_string in
      let _ = close_out outchan in
      compile_file ~redirect:true source_file_name [])
      (fun () -> Sys.remove source_file_name)

let compile_and_run : string -> string = 
  (* compiles Ikra source_code and returns its output as a string. *)

  fun source_code ->
    compile_code source_code
    (* let ruby_script = compile_code source_code *)
    (* in Util.run_ruby [ruby_script] *)

(* TODO: let this command accept options like -r (compile-and-run) and
   -v (verobse) *)

open Arg

let trace_typeinf : bool ref          = ref false
let generate_only : bool ref          = ref false
let conf_file     : string option ref = ref None
let backend       : string ref        = ref "cpp"
let target_dir    : string option ref = ref None
let ruby_file     : string option ref = ref None
let app_args      : string list ref   = ref []

let sors ors = String (fun s -> ors := (Some s))

let print_version : unit -> unit = fun _ ->
  (print_string Version.version;
   print_string "\n";
   exit 0)

let anon_arg a = match a, !ruby_file with
  | _, None   -> ruby_file := (Some a)
  | _, Some _ -> app_args := a::(!app_args)

let arg_specs = [
  "--trace-typeinf", Set trace_typeinf,
                " Show typeinference processes";
  "--generate",      Set generate_only,
                " Generate but not build CXX and Ruby programs";
  "--gen",           Set generate_only,
                " Generate but not build CXX and Ruby programs";
  "--conf",          sors conf_file,
                "[pathname] Specify a configuration file";
  "--backend",       Set_string backend, 
                "[cpp|cuda] Specify a backend (default: cpp)";
  "--target-dir",    sors target_dir,   
                "[pathname] Specify a directory to store generated programs";
  "--version",       Unit print_version, 
                " Print the version number and exit";
  "--",              Rest anon_arg, 
                "[arguments...] Pass to the Ruby program";
]


let usage = "a compiler for Ikra, a data-parallel dialect of Ruby.  Options: "

let printer = fun (c,s) -> 
  Printf.sprintf "c=%s, s=[%s]" (Cresolve.set_print c)
    (String.concat ","
       (List.map (fun (t1,t2) ->
         (Ikra_type.string_of t1) ^ "->" ^ (Ikra_type.string_of t2))
          s))

let parse_and_compile : ?redirect:bool -> string array -> 'a = 
  fun ?(redirect=false) argv ->
  try 
    Arg.parse_argv ~current:(ref 0) argv (align arg_specs) anon_arg usage;
    (* Printf.printf "\nruby file = %s, app_args =%s\n" *)
    (*   (Util.Option.get !ruby_file) (String.concat ";" !app_args); *)
    Config.initialize !conf_file;
    if !trace_typeinf
    then ((Cresolve.trace :=
             fun cs -> print_string (printer cs); print_string "\n");
          Cresolve.show_trace := true)
    else ();
    compile_file
      ~redirect
      ~cleanup:false
      ~generate_only:!generate_only
      ~backend:!backend
      ~target_dir:!target_dir
      (Util.Option.get(!ruby_file))
      (List.rev(!app_args))
  with
    | Arg.Help msg
    | Arg.Bad  msg -> output_string stderr msg; msg

let main _ = parse_and_compile Sys.argv

(* let rec getopts_and_compile ~generate_only ~conf_file ~backend ~target_dir *)
(*     cursor = *)
(*   match Sys.argv.(cursor) with *)

(*     | "--trace-typeinf" -> *)
(*       ((Cresolve.trace := *)
(*           fun cs -> print_string (printer cs); print_string "\n"); *)
(*        getopts_and_compile ~generate_only ~conf_file ~backend ~target_dir *)
(*          (cursor+1)) *)

(*     | "--gen" | "--generate" -> *)
(*       getopts_and_compile ~conf_file ~backend ~target_dir *)
(*         ~generate_only:true (cursor+1) *)

(*     | "--conf" -> *)
(*       getopts_and_compile ~generate_only ~backend ~target_dir *)
(*         ~conf_file:(Some(Sys.argv.(cursor+1))) (cursor+2) *)

(*     | "--backend" -> *)
(*       getopts_and_compile ~generate_only ~conf_file ~target_dir *)
(*         ~backend:Sys.argv.(cursor+1) (cursor+2) *)

(*     | "--target_dir" | "-t" -> *)
(*       getopts_and_compile ~generate_only ~conf_file ~backend *)
(*         ~target_dir:(Some (Sys.argv.(cursor+1))) (cursor+2) *)
        
(*     | filename -> *)
(*       Config.initialize conf_file; *)
(*       compile_file ~cleanup:false ~generate_only ~backend ~target_dir *)
(*         filename *)

(* let dummy _ = getopts_and_compile *)
(*   ~generate_only:false *)
(*   ~conf_file:None *)
(*   ~backend:"cpp" *)
(*   ~target_dir:None *)
(*   1 *)

