type t =
  | Tint			 (* integers *)
  | Tfloat			 (* floating point numbers *)
  | Tbool			 (* Boolean *)
  | Trange                       (* a pair (or triple) of numbers *)
  | Tarray of int * t		 (* (n,t) n-dimensional array of t  *)
  | Tneighbor of int * int * t   (* (m,n,t) m-neighbor expanded Tarray(n,t) *)
  | Tobject			 (* dynamic type *)
  | Tvar of int			 (* type variable distinguished by numbers *)
  | Tclass of string		 (* class objects *)
  | Ttuple of t list             (* tuples *)
  (* methods and argument blocks, where Tproc([t1;..;tn], t0, Some
     (Tproc t)) denotes a method that receives n parameters of type
     t1..tn, returns values of type t0, and takes a block argument of
     type Tproc t. *)
  | Tproc of t list * t * t option

let rec string_of : t -> string =
  function
    | Tint -> "int"
    | Tfloat -> "float"
    | Tbool -> "bool"
    | Trange -> "range"
    | Tarray(dim,base) ->
      Printf.sprintf("array(%d,%s)") dim (string_of base)
    | Tneighbor(n,dim,base) ->
      Printf.sprintf("neighbor(%d,%d,%s)") n dim (string_of base)
    | Tobject -> "object"
    | Tvar(n) -> "t"^(string_of_int n)
    | Tclass(c) -> "class<"^(c)^">"
    | Ttuple(typs) ->
      "tuple<"^(String.concat "," (List.map string_of typs))^">"
    (* t1*...*tn->t0&tb *)
    | Tproc(argument_types, return_type, block) ->
      (String.concat "*" (List.map string_of argument_types))
      ^ "->"
      ^ (string_of return_type)
      ^ (match block with | Some bt -> "&"^(string_of bt) | None -> "")

let type_order = function
  | Tint -> 0
  | Tfloat -> 1
  | Tbool -> 2
  | Tarray _ -> 3
  | Tobject _ -> 4
  | Tvar _ -> 5
  | Tclass _ -> 6
  | Ttuple _ -> 7
  | Tneighbor _ -> 8
  | Trange -> 9
  | Tproc _ -> 10

let rec compare_types x y =
  let rec list_compare = function
    | ([],[])   -> 0
    | ([],_::_) -> -1
    | (_::_,[]) -> 1
    | (h1::t1,h2::t2) -> let c = compare_types h1 h2 in
                         if c=0 then list_compare (t1,t2)
                         else c in
  match x,y with
    | x,y when x=y -> 0
    | x,y when type_order x < type_order y -> -1
    | x,y when type_order x > type_order y -> 1
    | Tarray(dim1,element1),Tarray(dim2,element2) when dim1=dim2 ->
      compare element1 element2
    | Tarray(dim1,element1),Tarray(dim2,element2) -> dim1 - dim2
    | Tvar(n1),Tvar(n2) -> n1 - n2
    | Tclass(c1),Tclass(c2) -> String.compare c1 c2
    | Ttuple(t1),Ttuple(t2) -> list_compare (t1,t2)
    | Tneighbor(m1,n1,t1),Tneighbor(m2,n2,t2) when m1=m2 && n1=n2 ->
      compare t1 t2
    | Tneighbor(m1,n1,t1),Tneighbor(m2,n2,t2) when m1=m2 -> n1 - n2
    | Tneighbor(m1,n1,t1),Tneighbor(m2,n2,t2) -> m1 - m2
    | Tproc(a1,r1,b1), Tproc(a2,r2,b2)
      when list_compare(a1,a2)=0 && r1=r2 -> (match (b1,b2) with
        | (Some bb1, Some bb2) -> compare_types bb1 bb2
        | (Some _,   None)     -> -1
        | (None,     Some _)   -> 1
        | (None,     None)     -> 0)
    | Tproc(a1,r1,_), Tproc(a2,r2,_)
      when list_compare(a1,a2)=0 -> compare_types r1 r2
    | Tproc(a1,r1,_   ), Tproc(a2,r2,_)
      -> list_compare(a1,a2)
    | _ -> invalid_arg ("dummy fallthrough: x="^ (string_of x) ^", y="^
                           (string_of y))


let compare = compare_types
