(* given a program, insert variable declarations at the beginnig of
   each block *)

open Ast
open Ast.Untyped_constructor 
open Ast.VarSet

type vset = VarSet.t

let pseudo_vars = List.fold_right add ["self"; Ast.implicit_self_name]
    empty

let let_intro : tast -> tast =

  (* to insert type_decl's for binding local variables into a given
     expression

     let_intro <<a=0>> -> <<type_decl(Obect,a); a=0>>
  *)

  (* ********************************************************************** *)
  let rec maybe_intro
      : tast option -> (vset * vset * (vset -> tast option)) =
    (* apply intro to optional expression *)
    
    function
      | Some exp -> let (fvs,gvs,exp') = intro exp in
                    (fvs, gvs, fun bound -> Some (exp' bound))
      | None     -> (empty, empty, fun _ -> None)

  and map_intro : tast list -> (vset * vset * (vset -> tast list)) =
    (* apply intro to a list of expressions *)

    fun exps ->
      let fvs_exps = List.map intro exps in
      (List.fold_left union empty (List.map (fun (fvs,_,_) -> fvs) fvs_exps),
       List.fold_left union empty (List.map (fun (_,gvs,_) -> gvs) fvs_exps),
       fun bound -> List.map (fun (_,_,exp) -> exp bound) fvs_exps)

  (* Consider

     intro { a =1; m(){ exp } } [] .

     It should eventually call

     intro exp [a]

     because a is lexically bound variable for exp.  This will
     require that the intro first collects bound variables at
     first round, and then go into inner block arguments.
  *)
  and intro : tast -> (vset * vset * (vset -> tast)) =
    (* given an expression, return a set of free local variables, a
       set of free global variables and a function that generates a
       let-inserted expression *)
    
    fun ((exp,_) as texp) -> match exp with

      | Local_var v when mem v pseudo_vars
          -> (empty, empty, fun _ -> local_var v) 
      | Local_var v -> (singleton v, empty, fun _ -> local_var v) 

      | Constant_assign(v,e) ->
        let (fvs, gvs, egen) = intro e in
        (fvs, add v gvs, fun bound -> constant_assign(v, egen bound))

      | Local_assign(v,e) ->
        let (fvs, gvs, egen) = intro e in
        (add v fvs, gvs, fun bound -> local_assign(v, egen bound))

      | If(cond,then_exp,else_exp) ->
        let (fvs1,gvs1, cond') = intro cond in
        let (fvs2,gvs2, then_exp') = intro then_exp in
        let (fvs3,gvs3, else_exp') = maybe_intro else_exp in
        (union (union fvs1 fvs2) fvs3,
         union (union gvs1 gvs2) gvs3,
         fun bound -> if_(cond' bound,then_exp' bound, else_exp' bound))

      | While(cond,body) ->
        let (fvs1,gvs1,cond') = intro  cond in
        let (fvs2,gvs2,body') = intro  body in
        (union fvs1 fvs2, union gvs1 gvs2,
         fun bound -> while_(cond' bound, body' bound))

      | Call(receiver, mname, args, block) ->
        let (fvs1, gvs1, receiver') = intro  receiver in
        let (fvs2, gvs2, args')     = map_intro  args in
        let (fvs3, gvs3, block')    = maybe_intro  block in
        ((union fvs1 fvs2)(* union (union fvs1 fvs2) fvs3 *),
         (union (union gvs1 gvs2) gvs3),
         fun bound ->
           call(receiver' bound, mname, args' bound, block' bound))

      (* NOTE: def(x1..xn) exps end ==>
         def(x1..xn) type_decl(fvs ...); exps end
         where vars are the free variables excluding {x1,..,xn} in
         exps
      *)
      | Method_def(mname,parameters,body) ->
        let (fvs,gvs,body') = intro body in
        (empty, gvs,
         fun bound ->
           let formals =
             List.fold_right add (variables_in_pattern parameters) empty in
           let body'' = body' (union (union bound formals) fvs) in
           method_def(mname,
                      parameters,
                      match List.map (fun v -> (v,Ikra_type.Tobject))
                        (elements (diff fvs formals)) with
                          | [] -> body''
                          | tdecl ->type_decl(tdecl, [body''])))

      | Block(bodies)  ->
        let (fvsb, gvsb, bodies') = map_intro bodies in
        (fvsb, gvsb, fun bound -> block(bodies' bound))
      | Proc(pattern,body)  ->
        let (fvsb, gvsb, body') = intro body in
        let formals =
          List.fold_right add (variables_in_pattern pattern) empty in
        (diff fvsb formals,
         gvsb,
         fun bound ->
           let body'' = body' (union bound fvsb) in
           let decls = (union formals (diff fvsb bound)) in
           proc(pattern,
                 match List.map (fun v -> (v,Ikra_type.Tobject))
                   (elements decls) with
                     | [] -> body''
                     | tdecl -> type_decl(tdecl,[body''])))

      | Yield(exp) ->
        let fvs,gvs,exp' = intro exp in
        (fvs, gvs, fun bound -> yield(exp' bound))

      | Array(exps) ->
        let fvs, gvs, exps' = map_intro  exps in
        (fvs, gvs, fun bound -> array(exps' bound))

      | For(var,range,body) ->
        let fvs1, gvs1, range' = intro range in
        let fvs2, gvs2, body' = intro body in
        (add var (union fvs1 fvs2),
         (union gvs1 gvs2),
         fun bound -> for_ (var,range' bound,body' bound))

      | Return(exp1) ->
        let fvs1, gvs1, exp1' = maybe_intro exp1 in
        (fvs1, gvs1, (fun bound -> return_o (exp1' bound)))

      (* type_decl (tdecl list, body) --- only after type inference *)
      | Type_decl _   -> invalid_arg
        ("this should not appear "^(string_of texp))

      (* Literal and Constant *)
      | Literal l -> (empty, empty, fun _ -> literal l)
      (* NOTE: a constant reference does not contribute to the global
         variable set.  This is because, constants that are not
         assigned in the program are not user defined. *)
      | Constant c -> (empty, empty, fun _ -> constant c) in

  (* the top level *)
  fun exp -> match intro exp with
    | (vars, gvars, exp') when is_empty vars && is_empty gvars
        -> exp' pseudo_vars
    | (vars, gvars, exp') (* when vars are not empty *) ->                      
      type_decl (List.map (fun v -> (v,Ikra_type.Tobject)) 
                   (elements (diff (union vars gvars) pseudo_vars)),
                 [exp' (union vars pseudo_vars)])
      
