(* ast with types *)

type parameter_pattern =
  | Var_pattern of string * Ikra_type.t
  | Array_pattern of parameter_pattern list


type ast =
  (* method call (receiver, method_name, operands, block) *)
  | Call of tast option * string * tast list * tast option
  (* literal (literal) *)
  | Literal of Ast.literal
  (* constant (name) *)
  | Constant of string 
  (* assignment to local variable (name,ast) *)
  | Local_assign of string * tast
  (* local variable reference (name) *)
  | Local_var of string
  (* block structure (variables, body)*)
  | Block of parameter_pattern * tast list
  (* method definition (method_name, variables, body) *)
  | Method_def of string * parameter_pattern * tast
  (* yield (ast) *)
  | Yield of tast
  (* array literal (ast list) *)
  | Array of tast list
  (* while loop (cond, body) *)
  | While of tast * tast 
  (* conditional branch (cond, then, else) *)
  | If of tast * tast * tast option
  (* for loop (var, range, body) *)
  | For of string * tast * tast

  (* type_decl (tdecl list, body) --- only after type inference *)
  | Type_decl of Ikra_type.t Ast.type_decl list * tast list
and tast = ast * Ikra_type.t

let maybe = Util.Option.maybe

let rec forget_parameter_pattern =
  function
    | Var_pattern(var,typ) -> Ast.Var_pattern var
    | Array_pattern patterns ->
      Ast.Array_pattern (List.map forget_parameter_pattern patterns)

let rec forget : tast -> 'a Ast.ast =
  
  function
    (* method call *)
    | Call (receiver, method_name, operands, block),_ ->
      Ast.Call (maybe forget receiver,
	    method_name,
	    List.map forget operands,
	    maybe forget block)
    (* literal  *)
    | Literal(literal),_ -> Ast.Literal literal
    (* constant  *)
    | Constant(name),_  -> Ast.Constant(name)
    (* assignment to local variable *)
    | Local_assign (name,ast),_  -> Ast.Local_assign (name,forget ast)
    (* local variable reference *)
    | Local_var (name),_  ->Ast.Local_var (name)
    (* block structure*)
    | Block (variables, body),_  ->
      Ast.Block (forget_parameter_pattern variables, List.map forget body)
    (* method definition *)
    | Method_def (method_name, variables, body),_  ->
      Ast.Method_def (method_name,
		      forget_parameter_pattern variables,
		      forget body)
    (* yield *)
    | Yield(exp),_  -> Ast.Yield(forget exp)
    (* array literal *)
    | Array(exps),_  -> Ast.Array(List.map forget exps)
    (* while loop *)
    | While (cond, body),_  -> Ast.While (forget cond, forget body)
    (* conditional branch *)
    | If (cond, then_exp, else_exp),_  ->
      Ast.If (forget cond, forget then_exp, maybe forget else_exp)
    (* for loop *)
    | For (var, range, body),_  ->
      Ast.For (var, forget range, forget body)
    (* type_decl --- only after type inference *)
    | Type_decl (tdecls, body),_ ->
      (* Ast.Block(Ast.Array_pattern[], List.map forget body) *)
      Ast.Type_decl(tdecls, List.map forget body)
      (* invalid_arg "unexpected type_decl"  *)


let string_of : ?hide_type_decl:bool -> tast -> string =
  fun ?(hide_type_decl=false) -> 
  function e -> Ikra_type.string_of_ast ~hide_type_decl:hide_type_decl (forget e)
let remove_first : ('a -> bool) -> 'a list -> ('a * 'a list) =
  fun test lst ->
    let rec loop = function
      | ([],_) -> raise Not_found
      | (hd::tl,prev) when test hd -> (hd, (List.rev prev) @ tl)
      | (hd::tl, prev) -> loop (tl, hd::prev) in
    loop(lst,[])

let return_type_of_body : tast list -> (Ikra_type.t * tast list) option =
  function 
    | (Type_decl(decls,body_exps),t)::rest_exps  ->
      (try let ((_,typ), other_decls) =
	     remove_first (fun (name, _) -> name = "_RETURN") decls in
	   Some(typ, (Type_decl(other_decls, body_exps),t)::rest_exps)
       with Not_found -> None)
    | _ -> None

let lookup_in_tdecl :
    string -> Ikra_type.t -> Ikra_type.t Ast.type_decl list -> Ikra_type.t =
  fun var default decls ->
    try List.assoc var decls
    with Not_found -> default

let replace_types : tast list -> parameter_pattern -> parameter_pattern =
  function
    | (Type_decl(decls, _),_)::_ ->
      let rec replace =
	function
	  | Var_pattern (v,t) -> Var_pattern(v, lookup_in_tdecl v t decls)
	  | Array_pattern pat -> Array_pattern (List.map replace pat) in
      replace
    | _ -> fun pat -> pat

let rec recognize_paratemer_pattern :
    Ast.parameter_pattern -> parameter_pattern = function
    | Ast.Var_pattern var -> Var_pattern (var, Ikra_type.Tobject)
    | Ast.Array_pattern parameters ->
      Array_pattern (List.map recognize_paratemer_pattern parameters)

let rec recognize : 'a Ast.ast -> tast =
  (* recognize ast constructs a typed AST with recognizing Ruby
     encoded types *)

  let convert = function
    (* method call *)
    | Ast.Call (receiver, method_name, operands, block) ->
      Call (maybe recognize receiver,
	    method_name,
	    List.map recognize operands,
	    maybe recognize block)
    (* literal  *)
    | Ast.Literal(literal) -> Literal literal
    (* constant  *)
    | Ast.Constant(name) -> Constant(name)
    (* assignment to local variable *)
    | Ast.Local_assign (name,ast) -> Local_assign (name,recognize ast)
    (* local variable reference *)
    | Ast.Local_var (name) ->Local_var (name)
    (* block structure*)
    | Ast.Block (variables, body) ->
      Block (recognize_paratemer_pattern variables, List.map recognize body)
    (* method definition *)
    | Ast.Method_def (method_name, variables, body) ->
      Method_def (method_name,
		  recognize_paratemer_pattern variables,
		  recognize body)
    (* yield *)
    | Ast.Yield(exp) -> Yield(recognize exp)
    (* array literal *)
    | Ast.Array(exps) -> Array(List.map recognize exps)
    (* while loop *)
    | Ast.While (cond, body) -> While (recognize cond, recognize body)
    (* conditional branch *)
    | Ast.If (cond, then_exp, else_exp) ->
      If (recognize cond, recognize then_exp, maybe recognize else_exp)
    (* for loop *)
    | Ast.For (var, range, body) ->
      For (var, recognize range, recognize body)
    (* type_decl --- only after type inference *)
    | Ast.Type_decl (tdecls, body)
      -> Type_decl (tdecls, List.map recognize body) in

  function
    (* Convert a pseudo function call to _t into a type of the
       expression.
       _t(exp,t) -> exp of type t
    *)
    | Ast.Call (None, "_t", [exp;texp], None) ->
      (convert exp, Ikra_type.of_ast texp)

    (* When a block has a _type_decl inside that specifies the type of
       _RETURN pesudo variable, use the type as the type of the block
       itself.
       { |v1,...| _type_decl(...,t, :_RETURN,...) ... } 
       -> { |v1,...| _type_decl(...,...) ... } of type t
    *)
    | Ast.Block _ as exp ->
      (match convert exp with
	| Block(variables, body_exps)
	    when Util.Option.is_some (return_type_of_body body_exps) ->
	  let typ,body_exps = Util.Option.get(return_type_of_body body_exps) in
	  let typed_vars = replace_types body_exps variables in
	  (Block(typed_vars, body_exps),typ)
	| e -> (e, Ikra_type.Tobject))
    | e -> (convert e,Ikra_type.Tobject)

