open OUnit2
open Ast

let _ = Config.initialize None

let testdata_expressions : (tast * tast) list =
  (* list of untyped and typed expressions, which serve as an example
     of input and output of type inferencing.  *)
  
  List.map
    (fun (input,output) ->
      (Test_util.parse_ruby_string input,
       Test_util.parse_ruby_string output))
    [
      "			       for a in 1..2 do p a; end",
      "_typedecl(Integer, :a); for a in 1..2 do p a; end";

      "def m(a,b)				    a+b+1 end",
      "def m(a,b) _typedecl(Integer, :a, :b); a+b+1 end";

      "Array.new(3){ |x|			       x*3 }",
      "Array.new(3){ |x| _typedecl(Integer, :x); x*3 }";

      "a.map{ |x|			       x*3 }",
      "_typedecl([Array,1,Integer], :a); a.map{ |x| _typedecl(Integer, :x); x*3 }";

      "			     a=0",
      "_typedecl(Integer, :a); a=0";

      "			   a=0.0",
      "_typedecl(Float, :a); a=0.0";

      "			       a=0.0; b=a",
      "_typedecl(Float, :a, :b); a=0.0; b=a";
    ]

let confirm_no_free_variables : 'a Env.env -> Ast.tast -> unit =
  fun _ exp ->
    assert_equal ~printer:Ast.string_of_varset
      Ast.VarSet.empty
      (Ast.free_variables exp)

(* TODO check if this test should be revived *)
(* let test_no_free_variables = *)
(*   (\* to confirm that there are no free variables in the result of type *)
(*      inferences. *\) *)
(*   List.map (fun (source,_) -> *)
(*     ("no_free_variables_in_"^(Ast.string_of (source))) >:: *)
(*       (fun _ -> *)
(*         confirm_no_free_variables Env.empty *)
(* 	   (Typeinf.typeinf source))) *)
(*     testdata_expressions *)

let testdata_type_decl_recognition : (tast * tast) list =
  let open Ast.Untyped_constructor in
  let open Ikra_type in
 [
  
  (* Typeinf.recognize_type_decls will convert the following expression *)
  (Block([Call((Local_var "self",Tobject),
	       "_typedecl",
		     [Constant("Integer"),Tobject;
		      Literal(Symbol_literal("a")),Tobject],
		     None),Tobject;
		Local_assign("a",(Literal(Integer_literal(123)),Tobject)),Tobject]),Tobject),
  (* into the following one, in which the Type_decl node represents
     the type declaration *)
  (Type_decl(["a",Ikra_type.Tint], 
	     [Local_assign("a",(Literal(Integer_literal(123)),Tobject)),
              Tobject]),
   Tobject);
  
  (Block([Call((Local_var "self",Tobject), "_typedecl",
	       [Constant("Integer"),Tobject;
		Literal(Symbol_literal("a")),Tobject;
		Literal(Symbol_literal("b")),Tobject],
	       None),Tobject;
	  (Local_assign("a",(Local_var("b"),Tobject)),Tobject)]), Tobject),
  (Type_decl(["a",Ikra_type.Tint; "b",Ikra_type.Tint], 
	     [Local_assign("a",(Local_var("b"),Tobject)),Tobject]),Tobject);


  (Method_def("m", Var_pattern("x",Tobject),
	      (Block([Call((Local_var "self",Tobject), "_typedecl",
			  [Constant("Integer"),Tobject;
			   Literal(Symbol_literal("x")),Tobject;
			   Literal(Symbol_literal("_RETURN")),Tobject],
			  None),Tobject;
		     Local_var("x"),Tobject;]),Tobject)),Tobject),
  (Method_def("m", Var_pattern("x",Tint),
	      (Type_decl(["x",Tint; "_RETURN", Tint;],
		         [Local_var("x"),Tobject;]),Tint)),
   Tobject);

  (* recognition of array types *)
  (Block([Call((Local_var "self",Tobject),
	       "_typedecl",
	       [Array[Constant("Array"),Tobject;
                      Literal(Integer_literal(1)),Tobject;
                      Constant("Integer"),Tobject;],Tobject;
		Literal(Symbol_literal("a")),Tobject],
	       None),Tobject;
	  Local_assign("a",(Literal(Integer_literal(123)),Tobject)),Tobject]),Tobject),
  (* recognized *)
  (Type_decl(["a",Ikra_type.Tarray(1,Tint)], 
	     [Local_assign("a",(Literal(Integer_literal(123)),Tobject)),
              Tobject]),
   Tobject);

  (* recognition of tuple types *)
  (Block([Call((Local_var "self",Tobject),
	       "_typedecl",
		     [Array[Constant("Tuple"),Tobject;
                            Constant("Integer"),Tobject;
                            Constant("Integer"),Tobject;],Tobject;
		      Literal(Symbol_literal("a")),Tobject],
		     None),Tobject;
		Local_assign("a",(Literal(Integer_literal(123)),Tobject)),Tobject]),Tobject),
  (* recognized *)
  (Type_decl(["a",Ikra_type.Ttuple[Tint;Tint]], 
	     [Local_assign("a",(Literal(Integer_literal(123)),Tobject)),
              Tobject]),
   Tobject);

(* recogniziton of procedure parameters *)
  (call (local_var "self", "m", [],
         Some (proc
                 (Var_pattern ("x",Tobject),
                  block[
                    call (local_var "self", "_typedecl",
                          [constant "Integer"; literal(Symbol_literal "x")],
                          None);
                    (local_var "x")
                  ])
         ))),
  (call (local_var "self", "m", [],
         Some (proc
                 (Var_pattern ("x",Tint),
                  type_decl (["x", Tint],
                             [
                               (local_var "x")
                             ]))
         )));
  
]

let test_type_decl_recognition =
  (* to check if recognize_type_decls can correctly convert a type
     declaration represented as a pseudo function call into a special
     purpose AST *)

  List.map (fun (input,output) ->
    ("type_decl_recognition("^ (Ast.string_of input) ^")") >::
      (fun _ ->
	let actual = (Typeinf.recognize_type_decls input) in
	assert_equal ~printer:(Ast.mlexp_of Ikra_type.string_of)
	  output actual))
    testdata_type_decl_recognition

let test_type_decl_recognition2 =
  (* to check if recognize_type_decls can convert the typed
     expressions in testdata_expressions *)

  List.map (fun (_, output) ->
    ("type_decl_recognition2("^ (Ast.string_of output) ^")") >::
      (fun _ ->
	let _ = Typeinf.recognize_type_decls output
	in ()))
    testdata_expressions
    
let test_type_decl_recognition3 _ =
  (* to confirm whether a top-level type declaration can be
     recognized *)
  let open Ikra_type in
  let exp =
    Test_util.parse_ruby_string "_typedecl(Integer, :a); a=0;# a=a+1" in
  assert_equal ~printer:(Ast.mlexp_of Ikra_type.string_of)
    (Type_decl([("a",Ikra_type.Tint)],
	       [Local_assign("a",(Literal(Integer_literal 0),Tobject)),
                Tobject]),
     Tobject)
    (Typeinf.recognize_type_decls exp)

let test_type_decl_recognition4 =
  "type_decl_recognition4">::
    (fun _ ->
      let before = (Test_util.parse_ruby_string
		      "PArray.new(10){ |x| _typedecl(Integer, :x); x }") in
      let recon = Typeinf.recognize_type_decls before in
      (match before with
	(* |Call(_,"new",_, *)
	(*        Some((Block((Call(_,"_typedecl",_,_),_)::_),_))),_ -> () *)
	|Call(_,"new",_,
	       Some(
                 Proc(_,
                      (Block(
                        (Call(_,"_typedecl",_,_),_)
                        ::_),
                      _)),
                 _)),_
          -> ()
	|Call(_,"new",_, Some((Block((Block _,_)::_),_))),_ ->
	  assert_failure "nested block"
	| _ -> assert_failure "no typedecl in before");
      match recon with
	| Call(_,"new",_,
	       Some(
                 Proc(_, (Type_decl _,_)),
                 _
               )
        ),_ -> ()
	| Call(_,"new",_,
	       Some(Block([Block _,_ as e]),_)),_ ->
	  assert_failure ("no decl but block:" ^ (Ast.string_of e))
	| Call(_,"new",_,
	       Some(Block((e)::_),_)),_ ->
	  assert_failure ("no decl but:" ^ (Ast.string_of e))
	(* | Call(_,"new",_, *)
	(*        Some(Block(_,(Call (_,m,_,_))::_))) -> *)
	(*   assert_failure ("no decl but call to:" ^ m) *)
	| Call(_,"new",_,
	       Some(Block(exps),_)),_ ->
	  assert_failure ("no decl in :" ^ 
			     (String.concat ";"
				(List.map Ast.string_of exps)))
	| _ -> assert_failure "not recognized"
    )

(* TODO check if this test is needed *)
(* let type_inf_tests = *)
(*   List.map (fun (input,expected) -> *)
(*     ("type inference test" ^ (Ast.string_of input))>:: *)
(*       (fun _ -> *)
(* 	assert_equal ~printer:Ast.string_of expected (Typeinf.typeinf input) *)
(*       )) testdata_expressions *)

let test_add_type_vars source =
  let rec assert_has_type_var : Ast.tast -> unit =
    function
      | exp,(Ikra_type.Tvar _) -> traverse exp
      | exp,t -> assert_failure
	((Ast.string_of (exp,t)) ^ " should have a type variable, but has a type of "
	 ^ (Ikra_type.string_of t))
  and for_all_has_type_var exps = List.for_all
      (fun exp -> assert_has_type_var exp; true) exps
  and maybe_has_type_var = function
      | Some exp -> assert_has_type_var exp
      | None -> () 
  and traverse =
    function
      | Call(receiver, _, args, block) ->
	assert_has_type_var receiver;
	ignore (for_all_has_type_var args);
	maybe_has_type_var block
      | Literal _ -> ()
      | Constant _ -> ()
      | Local_assign(v,exp)
      | Constant_assign(v,exp) -> assert_has_type_var exp
      | Local_var _ -> ()
      | Block(bodies) -> ignore (for_all_has_type_var bodies)
      | Proc(param,body) -> assert_has_type_var body
      | Method_def(_,_,body) -> assert_has_type_var body
      | Yield(arg) -> assert_has_type_var arg
      | Array(exps) -> ignore (for_all_has_type_var exps)
      | While(cond,body) -> assert_has_type_var cond; assert_has_type_var body
      | If(cond,then_exp,else_exp) ->
	assert_has_type_var cond;
	assert_has_type_var then_exp;
	maybe_has_type_var else_exp
      | For(_, range, body) ->
	assert_has_type_var range; assert_has_type_var body
      | Return(exp1) -> maybe_has_type_var exp1
      | Type_decl(_,exps) -> ignore (for_all_has_type_var exps)

  in
  ("test_add_type_vars("^source^")")>::
    (fun _ ->
      assert_has_type_var 
	(Typeinf.add_type_vars
	   (Test_util.parse_ruby_string source)))

let set_print = fun s -> "{" ^ 
  (String.concat "," (List.map Type_constraints.Constr.string_of
			(Type_constraints.Constraints.elements s))) ^ 
  "}"
let pp_diff formatter (exp,real) =
  let lacking = Type_constraints.Constraints.diff exp real in
  let extra =Type_constraints.Constraints.diff real exp in
  Format.fprintf formatter  "lacking:%s extra:%s" (set_print lacking) (set_print extra)

let test_gather source source_with_type_expected expected =
  ("gather:"^source)>:: (fun _->
  let source_exp = Test_util.parse_ruby_string source in
  let source_with_tdecl = Let_intro.let_intro source_exp in
  let source_with_tvar = Typeinf.add_type_vars source_with_tdecl in
  let source_str = Ast.string_of  ~show_type:true source_with_tvar in
  let constrs : Type_constraints.Constraints.t =
    Typeinf.generate_constraints  source_with_tvar in
    assert_equal ~printer:(fun s->s) source_with_type_expected source_str;
    assert_equal ~cmp:Type_constraints.Constraints.equal
      ~printer:set_print
      ~pp_diff
      expected constrs)

module ExpectedTypeConstraintConstructors = struct
  open Ikra_type
  let (<) t0 t1 = Type_constraints.Constr.Less_than(t0,t1)
  let (=) t0 t1 = Type_constraints.Constr.Equal(t0,t1)
  let (=>) t0 t1 = Tproc(t0,t1,None)
  let t n = Tvar n
  let cset =
    List.fold_left (fun set con -> Type_constraints.Constraints.add con set)
      Type_constraints.Constraints.empty
  let mtype receiver name args return =
    Type_constraints.Constr.Has_method(receiver,name,Tproc(args,return,None))
  let mtype' receiver name args (barg,bret) return =
    Type_constraints.Constr.Has_method
      (receiver,name,Tproc(args,return,Some(Tproc(barg,bret,None))))
  let mtype'' receiver name args blk return =
    Type_constraints.Constr.Has_method
      (receiver,name,Tproc(args,return,Some(blk)))
  let def receiver name args return =
    Type_constraints.Constr.Defines_method
      (receiver,name,Tproc(args,return,None))
end

let test_gather_cases =
  let open ExpectedTypeConstraintConstructors in
  let open Ikra_type in
  [
    test_gather "x=123"
      "(_typedecl(t3,:x);(x = ((123):t2)):t1):t0"
      (cset [ t 2 = Tint; t 2 < t 3; t 1 = t 3; t 1 = t 0; ]);
    test_gather "x=0.123"
      "(_typedecl(t3,:x);(x = ((0.123000):t2)):t1):t0"
      (cset [ t 2 = Tfloat; t 2 < t 3; t 1 = t 3; t 1 = t 0; ]);
    test_gather "x=if x then x else x end"
      "(_typedecl(t6,:x);(x = ((if (x):t5 then (x):t4 else (x):t3 end):t2)):t1):t0"
      (cset [ t 1 = t 0; t 1 = t 6; t 2 < t 6; t 3 < t 2;
    	      t 4 < t 2; t 5 < Tbool; t 3 = t 6; t 4 = t 6; t 5 = t 6 ]);
    test_gather "x=(while x do x; end)"
      "(_typedecl(t5,:x);(x = ((while (x):t4;(x):t3; end):t2)):t1):t0"
      (cset [ t 1 = t 0; t 1 = t 5; t 2 < t 5; t 4 < Tbool;
    	    t 2 = Tbool; t 3 = t 5; t 4 = t 5 ]);
    test_gather "x=1+1"
      "(_typedecl(t5,:x);(x = (((1):t4.+((1):t3)):t2)):t1):t0"
      (cset [ t 2 < t 5;                (* x:t5 = e:t2 *)
              t 4 = Tint;               (* 0:t4 *)
              t 3 = Tint;               (* 0:t3 *)
              mtype (t 4) "+" [t 3] (t 2); (* (e:t4 .+ e:t3):t2 *)
              t 1 = t 5;                   (* (x:t5 = e):t1 *)
              t 1 = t 0; ]);

    (* a method call with a block *)
    test_gather "x=x.m(123){ |y| 456 }"
      "(_typedecl(t10,:x);(x = (((x):t9.m((123):t8){( | y:t7 | (_typedecl(t6,:y);(456):t5):t4):t3}):t2)):t1):t0"
      (cset [ t 2 < t 10;               (* x:t10 = e:t2 *)
              t 9 = t 10;              (* x:t10 |- x:t9 *)
              t 8 = Tint;               (* 122:t8 *)
              mtype'' (t 9) "m" [t 8] (t 3) (t 2);
              t 5 = Tint;               (* 455:t5 *)
              t 5 = t 4;
              t 7 = t 6;                (* { |y:t7| _typedecl(t6,:y); ...} *)
              t 3 = ([t 7] => (t 4));
              t 1 = t 0;
              t 1 = t 10;               (* x:t10 |- (x = e):t1 *)
      ] );

    (* a method call with a block, in which a local variable is declared *)
    test_gather "x=x.m(123){ |y| z=y; 456 }"
      "(_typedecl(t14,:x);(x = (((x):t13.m((123):t12){( | y:t11 | (_typedecl(t9,:y,t10,:z);((z = ((y):t7)):t6;(456):t8):t5):t4):t3}):t2)):t1):t0"
      (cset [ t 2 < t 14;               (* x:t14 |- x = (e:t2) *)
              t 7 < t 10;               (* z:t10 |- z = (y:t7) *)
              t 8 < t 5;                (* e-1:t-1;e0:t8 ==> (e-1;e0):t8 *)
              t 1 = t 0;                (* e-1:t-1;e0:t8 ==> (e-1;e0):t8 *)
              t 1 = t 14;               (* x:t14 |- (x = e):t14 *)
              t 3 = ([t 11] => t 4);    (* { |y:t11| e:t14} : t11->t14 *)
              t 5 = t 4;                (* e-1:t-1;e0:t0 ==> (e-1;e0):t0 *)
              t 6 = t 10;               (* x:t14 |- (x = e):t14 *)
              t 7 = t 9;               (* x:t14 |- (x = e):t14 *)
              t 8 = Tint;
              t 11 = t 9;
              t 12 = Tint;
              t 13 = t 14;
              mtype'' (t 13) "m" [t 12] (t 3) (t 2);
      ] );

    (* a method call with a block, in which a local variable is declared *)
    test_gather "x=x.m(123){ |y,w| z=y; 456 }"
      "(_typedecl(t16,:x);(x = (((x):t15.m((123):t14){( | y:t12,w:t13 | (_typedecl(t9,:w,t10,:y,t11,:z);((z = ((y):t7)):t6;(456):t8):t5):t4):t3}):t2)):t1):t0"
      (cset [
	t 15 = t 16;
        mtype'' (t 15) "m" [t 14] (t 3) (t 2);
        t 14 = Tint;
        t 7 = t 10;
        t 7 < t 11;
        t 6 = t 11;
        t 8 = Tint;
        t 8 < t 5;
        t 5 = t 4;
        t 3 = ([t 12; t 13] => (t 4));
        t 2 < t 16;
        t 1 = t 16;
        t 1 = t 0;
        t 13 = t 9;
        t 12 = t 10;
      ] );

    test_gather "x=Array"
      "(_typedecl(t3,:x);(x = ((Array):t2)):t1):t0"
      (cset [t 2 < t 3; t 1 = t 3; t 1 = t 0; 
    	     t 2 = Tclass "Array"]);

    test_gather "x=Array.new(10){ |x| 0 }"
      "(_typedecl(t10,:x);(x = (((Array):t9.new((10):t8){( | x:t7 | (_typedecl(t6,:x);(0):t5):t4):t3}):t2)):t1):t0"
      (cset [
        t 2 < t 10;
        t 1 = t 10;
        t 9 = Tclass "Array";
        mtype'' (t 9) "new" [t 8] (t 3) (t 2);
        t 8 = Tint;
        t 3 = ([t 7] => (t 4));
        t 7 = t 6;
        t 5 = Tint;
        t 5 = t 4;
        t 1 = t 0;

      ]);

    test_gather "x=x[0]"
      "(_typedecl(t5,:x);(x = (((x):t4.[]((0):t3)):t2)):t1):t0"
      (cset [
        t 2 < t 5;
        t 4 = t 5;
        mtype (t 4) "[]" [t 3] (t 2);
        t 1 = t 5;
        t 1 = t 0;
        t 3 = Tint;
      ]);

    (* TODO: support a block without a parameter *)
    (* test_gather "x=Array.new(10){ 0 }" *)
    (*   "((_typedecl(t11,:x);(x = (((Array):t10.new((10):t9){( | x:t8 | (_typedecl(t7,:x);(0):t6):t5):t4}):t3)):t2):t1):t0" *)
    (*   (cset [ *)
    (* 	t 3 < t 11; t 1 = t 0; t 2 = t 1; t 2 = t 11; t 6 = Tint; *)
    (* 	t 6 = t 5; t 8 = t 7; t 9 = Tint; t 10 = Tclass "Array"; *)
    (* 	mtype' (t 10) "new" [t 12] ([t 8], t 5) (t 3) *)
    (*   ]); *)


    test_gather "def m(x) x*1; end"
      "(def m(x:t5)(((x):t4.*((1):t3)):t2):t1; end):t0"
      (cset [ def Tobject "m" [t 5] (t 1);
	      t 4 = t 5;
	      t 3 = Tint;
	      mtype (t 4) "*" [t 3] (t 2);
	      t 2 < t 1;
	    ]);

    test_gather "def m(x) x*1; end; m(10)"
      "((def m(x:t6)(((x):t5.*((1):t4)):t3):t2; end):t1;(m((10):t8)):t7):t0"
      (cset [ def Tobject "m" [t 6] (t 2); 
              mtype (t 5) "*" [t 4] (t 3);
              t 7 < t 0;
              t 3 < t 2;
              t 4 = Tint;
              t 5 = t 6;
              t 8 = Tint;
              t 9 = Tobject;
              mtype (t 9) "m" [t 8] (t 7);
	    ]);

    test_gather "a=a.map{ |(x,y),z| 0 }"
      "(_typedecl(t13,:a);(a = (((a):t12.map(){( | (x:t9,y:t10),z:t11 | (_typedecl(t6,:x,t7,:y,t8,:z);(0):t5):t4):t3}):t2)):t1):t0"
      (cset [ mtype'' (t 12) "map" [] (t 3) (t 2);
              t 3 = ([Ttuple[t 9; t 10]; t 11] =>  (t 4));
              t 2 < t 13;               (* a:t13 = exp:t2 *)
              t 1 = t 0;                (* (exp0; exp1:t1):t0 *)
              t 1 = t 13;               (* (a:t13 = exp):t1 *)
              t 5 = Tint;               (* -1:t5 *)
              t 5 = t 4;                (* ( ...; exp:t5):t4 *)
              t 9 = t 6; (* { |(x:t9,_),_| _typedecl(t6:x,...) ..} *)
              t 10 = t 7; (* { |(_,y:t10),_| _typedecl(..t7:y..) ..} *)
              t 11 = t 8; (* { |(__ ),z:t11| _typedecl(..t8:z..) ..} *)
              t 12 = t 13;              (* _typedecl(t13,); ...(a):t12... *)
              
            ]);


    test_gather "c=0..c"
      "(_typedecl(t6,:c);(c = ((((0):t3)..((c):t4)):t2)):t1):t0"
      (cset [
        t 3 = Tint;
        t 4 = t 6;
        mtype (t 5) "__PSEUDO_dot2" [t 3; t 4] (t 2);
        t 5 = Tobject;
        t 2 < t 6;
        t 1 = t 6;
        t 1 = t 0;
      ] );

    test_gather "for i in 1..10 do; i; end"
      "(_typedecl(t7,:i);(for i in (((1):t4)..((10):t5)):t3 do (i):t2 end):t1):t0"
      (cset [
        t 7 = Tint;
        t 3 < Trange;
        t 1 = t 3;
        t 1 = t 0;
        t 2 = Tint;
        t 4 = Tint;
        t 5 = Tint;
        mtype (t 6) "__PSEUDO_dot2" [t 4;t 5] (t 3);
        t 6 = Tobject;
      ] );

    test_gather "C=1"
      "(_typedecl(t3,:C);(C = ((1):t2)):t1):t0"
      (cset [
        t 2 < t 3;
        t 1 = t 3;
        t 1 = t 0;
        t 2 = Tint;
      ]);

    test_gather "def m(n) if true then return 0.5; end; n end"
      "(def m(n:t7)((if (true):t5 then (return (0.500000):t4):t3 end):t2;(n):t6):t1; end):t0"
      (cset [
        t 6 = t 7;
        t 5 = Tbool;
        t 5 < Tbool;
        t 4 = Tfloat;
        t 4 < t 1;
        t 6 < t 1;
        t 2 = Tobject;
        def Tobject "m" [t 7] (t 1);
      ]);

    test_gather "unknownMethod(){ 1 }"
      "(unknownMethod(){((1):t2):t1}):t0"
      (cset [
        t 2 = Tint;
        t 3 = Tobject;
        mtype'' (t 3) "unknownMethod" [] (t 1) (t 0);
        t 1 = ([] => (t 2));
      ]);

    test_gather "unknownMethod(){ |x| 1 }"
      "(unknownMethod(){( | x:t5 | (_typedecl(t4,:x);(1):t3):t2):t1}):t0"
      (cset [
        t 3 = Tint;
        t 3 = t 2;
        t 5 = t 4;
        t 6 = Tobject;
        mtype'' (t 6) "unknownMethod" [] (t 1) (t 0);
        t 1 = ([t 5] => (t 2))
      ]);

    test_gather "unknownMethod(){ |x| anotherUknown(){ 1 } }"
      "(unknownMethod(){( | x:t8 | \
        (_typedecl(t7,:x);(anotherUknown(){((1):t5):t4}):t3):t2):t1}):t0"
      (cset [
        t 5 = Tint;
        t 6 = Tobject;
        t 4 = ([] => (t 5));
        mtype'' (t 6) "anotherUknown" [] (t 4) (t 3);
        t 3 = t 2;
        t 8 = t 7;
        t 9 = Tobject;
        t 1 = ([t 8] =>(t 2));
        mtype'' (t 9) "unknownMethod" [] (t 1) (t 0);
      ]);

    test_gather "unknownMethod(){ 123 }"
      "(unknownMethod(){((123):t2):t1}):t0"
      (cset [
        mtype'' (t 3) "unknownMethod" [] (t 1) (t 0);
        t 3 = Tobject;
        t 2 = Tint;
        t 1 = ([] => (t 2));
        
      ]);

    test_gather  "if  false && true then 1 else 0 end"
    "(if (((false):t5)&&((true):t4)):t3 then (1):t2 else (0):t1 end):t0"
    (cset [
      t 1 < t 0;
      t 2 < t 0;
      t 1 = Tint;
      t 2 = Tint;
      t 4 = Tbool;
      t 5 = Tbool;
      t 3 < Tbool;
      mtype (t 5) "&&" [t 4] (t 3);
    ]);


    test_gather "[1, 1.0]"
      "([(1):t1,(1.000000):t2]):t0"
      (cset [
        t 1 = Tint;
        t 2 = Tfloat;
        t 0 = Ttuple([t 1;t 2]);
      ]);

  ]

let test_type_comparison =
  let open ExpectedTypeConstraintConstructors in 
  let open Ikra_type in [
    ("compare_def_constraints1">::
        fun _ ->
          assert_equal false
            (Type_constraints.Constraints.equal 
               (Type_constraints.Constraints.singleton
                  (def Tobject "m" [t 7] (t 8)))
               (Type_constraints.Constraints.singleton
                  (def Tobject "m" [t 7] (t 1)))));
    
    ("compare_def_constraints2">::
        fun _ ->
          assert_equal true
            (Type_constraints.Constraints.equal 
               (Type_constraints.Constraints.singleton
                  (def Tobject "m" [t 7] (t 8)))
               (Type_constraints.Constraints.singleton
                  (def Tobject "m" [t 7] (t 8)))));
]

let suite =
  "suite">:::
    (* test_no_free_variables @ *)
    test_type_decl_recognition @
    test_type_decl_recognition2 @
    ["test_type_decl_recognition3">::test_type_decl_recognition3] @
    [test_type_decl_recognition4;
    ] @
    (* type_inf_tests @ *)
    [test_add_type_vars
    "x.m(x,0,:a,x){ x }; x=[1,1]; def m(y) yield(y) end; while x do x end; \
	for x in x..x do x end"
    ] @
    test_gather_cases @
    test_type_comparison
;;

let () =
  run_test_tt_main suite
;;
