(* ultility functions shared by all modules *)

let read_lines_list : string -> string list =
  (* returns a list of strings in the given filename *)

  fun filename ->
    let inc = open_in filename in
    let rec loop lines =
	 try loop (input_line inc::lines)
	 with End_of_file -> 
	   close_in inc; (List.rev lines) in
    loop []


let read_lines : string -> string =
  (* returns a string in the given filename *)

  fun filename ->
    String.concat "\n" (read_lines_list filename)


let with_finally : (unit -> 'a) -> (unit -> unit) -> 'a =
  (* (with_finally try_proc finally_proc) runs try_proc and then
     finally_proc.  It ensures that the finally_proc is run even
     try_proc raises an exception. *)
  fun try_proc finally_proc ->
    try let ans = try_proc () in
	finally_proc (); ans
    with expn -> finally_proc (); raise expn

let with_temp_file : string -> string -> string -> (string -> 'a) -> 'a =
  (* (with_temp_file prefix suffix body proc) writes the given body
     into a temporary file, and then runs the proc with the filename
     of the temporary file.  The temporary filename is generated with
     the given prefix and suffix.  If prefix is a relative path name,
     the temporary file is created in a temporary directory. *)
  fun prefix suffix string proc ->
    let temp_file_name, outchan =
      if (Filename.is_relative prefix)
      then Filename.open_temp_file prefix suffix
      else (Filename.open_temp_file ~temp_dir:(Filename.dirname prefix)
              (Filename.basename prefix) suffix) in
    let _ = output_string outchan string in
    let _ = close_out outchan in
    with_finally (fun () -> proc temp_file_name)
      (fun () -> Sys.remove temp_file_name)

let make_temp_dir : string -> string =
  (* make_temp_dir base creates a directory in a temporary directory
     and returns the pathname of the directory.  The directory has a
     name based on the given base.*)

  fun base ->
    let name = Filename.temp_file base "" in
    (Sys.remove name; Unix.mkdir name 0o755; name)
      
let with_cd : string -> (unit -> 'a) -> 'a =
  (* (with_cd dir proc) evaluates the proc under the dir as CWD.
     After finished, CWD is restored as before.  NOTE: raise of an
     exception does not restore the CWD. *)

  fun dir proc -> 
    let cwd = Sys.getcwd () in
    Sys.chdir dir;
    with_finally proc
      (fun () -> Sys.chdir cwd)

(* let path_to_ruby = *)
(*   (\* an absolute path to the Ruby interpreter. it can use wildcards *)
(*      like tilda and asterisk *\) *)
(*   fun _ -> *)
(*     List.find Sys.file_exists  *)
(*       ["/home/masuhara/.rvm/bin/ruby-1.9.3-head"; *)
(* 			(\* rvm use ruby-1.9.3-head beforehand *\) *)
(*        "/home/masuhara/.rvm/bin/ruby-1.9.3-p448" ] *)

(* External_command_error(cmd,out,err) represents a non-zero exit of
   execution of cmd, with out and err are the outputs to the STDOUT
   and STDERR. *)
exception External_command_error of string * string * string

let run_command : ?redirect:bool -> string list -> string = 
  (* (run_command ~redirect:r command_and_args) invokes the given
     command_and_args, and returns its output as a string when r is
     true.  When the exit state of the command is other than zero, it
     raises External_command_error exception.  When r is false, the
     outputs are sent to the console. *)
  fun ?(redirect=true) command_and_args ->
    let body redirect_spec =
      let cmd = (String.concat " " (command_and_args @ redirect_spec)) in
      (Sys.command cmd,cmd) in
    if redirect then
      (with_temp_file "command" ".out" "" (fun stdout_file ->
        with_temp_file "command" ".out" "" (fun stderr_file ->
	  let (status,cmd) = body [">";stdout_file; "2>";stderr_file] in
	  if status != 0 then
            raise (External_command_error
		     (cmd,
		      read_lines stdout_file,
		      read_lines stderr_file))
	  else read_lines stdout_file)))
    else let status,cmd = body [] in
         if status != 0 then
           raise (External_command_error
                    (cmd,
                     "not recorded",
                     Printf.sprintf "error status %d" status ))
         else ""
           
let run_ruby : ?redirect:bool -> string list -> string =
  (* to run a Ruby interpreter with the given parameters, and returns
     its output as a string.  When redirect is true, the output from
     the interpreter is returned as a string, or an exception.
     Otherwise, the outputs are send to the console. *)

  fun ?(redirect=true) parameters ->
    run_command ~redirect ((Config.path_to_ruby ())::parameters)


let in_paren : string -> string =
  (* to compute a string surrounded with parentheses *)

  fun s -> "(" ^ s ^ ")"

module Option = struct

  let maybe : ('a -> 'b) -> 'a option -> 'b option =
    fun f -> function
      | Some v -> Some (f v)
      | None -> None
	
  let is_some = function
    | Some _ -> true
    | None -> false

  exception No_value
  let get = function
    | Some x -> x
    | None -> raise No_value

end

let rec take : int -> 'a list -> 'a list =
  (* take n lst returns a list of the first n elements in lst. *)
  function
    | 0 -> fun _ -> []
    | n -> 
      (function elm::lst -> elm::(take (n-1) lst)
        | _ -> raise (Failure "take"))

let rec drop : int -> 'a list -> 'a list =
  (* take n lst returns a list of the first n elements in lst. *)
  function
    | 0 -> fun lst -> lst
    | n -> 
      (function _::lst -> (drop (n-1) lst)
        | _ -> raise (Failure "drop"))

let mapi : (int -> 'a -> 'b) -> 'a list -> 'b list =
  (* map with index; mapi f [v0;..;vN] = [f 0 v0; f 1 v1; ...; f N vN] *)
  fun f ->
    let rec loop n = function
      | [] -> []
      | hd::tl -> (f n hd)::(loop (n+1) tl) in
    loop 0 

let rec split_last : 'a list -> 'a list * 'a = function
  | [last] -> ([],last)
  | hd::tl -> let butlast, last = split_last tl in
	      hd::butlast, last
  | _ -> invalid_arg "empty"
